/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "auxFunctions.h"

#include <standardFiles/typedefs.h>
#include <standardFiles/functions.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <atmosphere/atmosphere.h>
#include <unitConversion/unitConversion.h>
#include <runtimeInfo/runtimeInfo.h>

#include <memory>
#include <vector>
#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <numeric>
#include <utility>  // make_pair

#include "../../polar.h"

using std::endl;
using std::vector;

Point getTrueCoordinates(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const size_t& sectionID) {
    Point TrueCoordinates;
    TrueCoordinates.xCoordinate = aLiftingSurface.origin.x() + aLiftingSurface.sections.at(sectionID).origin.x();
    TrueCoordinates.yCoordinate = aLiftingSurface.origin.y() - aLiftingSurface.sections.at(sectionID).origin.z();
    TrueCoordinates.zCoordinate = aLiftingSurface.origin.z() - aLiftingSurface.sections.at(sectionID).origin.y();
    return TrueCoordinates;
}

Point get_l_mu_ref_Point(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface) {
    // get LE origin point of the wing
    double MAC_spanPosition = - geom2::measure::mean_aerodynamic_chord_position(aLiftingSurface);
    Point l_mu_ref_Point = getTrueCoordinates(aLiftingSurface, 0);
    geom2::Point_3 offset_LE_Point_CGAL = geom2::measure::offset_LE(aLiftingSurface, geom2::measure::mean_aerodynamic_chord_position(aLiftingSurface));
    Point TRUE_offset_LE_Point;
    TRUE_offset_LE_Point.xCoordinate = offset_LE_Point_CGAL.x();
    TRUE_offset_LE_Point.yCoordinate = - offset_LE_Point_CGAL.z();
    TRUE_offset_LE_Point.zCoordinate = - offset_LE_Point_CGAL.y();
    // add LE point at MAC to ref Point
    l_mu_ref_Point.xCoordinate += TRUE_offset_LE_Point.xCoordinate;
    l_mu_ref_Point.yCoordinate += TRUE_offset_LE_Point.yCoordinate;
    l_mu_ref_Point.zCoordinate += TRUE_offset_LE_Point.zCoordinate;
    // add quarter chord to ref Point
    double MAC_Twist = getLocalTwist(aLiftingSurface, MAC_spanPosition);
    double MAC_length = getLocalChordLength(aLiftingSurface, MAC_spanPosition);
    l_mu_ref_Point.xCoordinate += 0.25 * MAC_length * cos(convertUnit(DEGREE, RADIAN, MAC_Twist));
    l_mu_ref_Point.zCoordinate += - 0.25 * MAC_length * sin(convertUnit(DEGREE, RADIAN, MAC_Twist));
    return l_mu_ref_Point;

}

double getCGAL_ZCoordinate(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    double CGAL_ZCoordinate = aLiftingSurface.origin.y() - yPosition;
    return CGAL_ZCoordinate;
}

bool checkIfWithinWing(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    double localZCoordinate = getCGAL_ZCoordinate(aLiftingSurface, yPosition);
    double innerZCoordinate = aLiftingSurface.sections.front().origin.z();
    double outerZCoordinate = aLiftingSurface.sections.back().origin.z();
    bool yPositionWithinWing = false;
    if (outerZCoordinate <= localZCoordinate || localZCoordinate <= innerZCoordinate) {
        yPositionWithinWing = true;
    }
    return yPositionWithinWing;
}

size_t determineWingSection(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    if (!checkIfWithinWing(aLiftingSurface, yPosition)) {
        myRuntimeInfo->err << "Given YPosition is outside the wing!" << endl;
        exit(1);
    }
    double localZCoordinate = getCGAL_ZCoordinate(aLiftingSurface, yPosition);
    size_t theInnerSectionID(0);
    vector<double> sectionCoordinates;
    for (size_t sectionID(0); sectionID < aLiftingSurface.sections.size(); ++sectionID) {
        sectionCoordinates.push_back(aLiftingSurface.sections.at(sectionID).origin.z());
    }
    for (size_t sectionID(0); sectionID < sectionCoordinates.size() - 1; ++sectionID) {
        if (sectionCoordinates.at(sectionID) >= localZCoordinate || localZCoordinate >= sectionCoordinates.at(sectionID + 1)) {
            theInnerSectionID = sectionID;
            break;
        }
    }
    return theInnerSectionID;
}

double getLocalChordLength(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    if (!checkIfWithinWing(aLiftingSurface, yPosition)) {
        myRuntimeInfo->err << "Given YPosition is outside the wing!" << endl;
        exit(1);
    }
    double localMinusZ = getCGAL_ZCoordinate(aLiftingSurface, yPosition);
    return geom2::measure::chord(aLiftingSurface, localMinusZ);
}

double getLocalTwist(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    if (!checkIfWithinWing(aLiftingSurface, yPosition)) {
        myRuntimeInfo->err << "Given YPosition is outside the wing!" << endl;
        exit(1);
    }
    double localCGAL_Z = getCGAL_ZCoordinate(aLiftingSurface, yPosition);
    return convertUnit(RADIAN, DEGREE, geom2::measure::twist(aLiftingSurface, localCGAL_Z));
}

double getLocalThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition) {
    if (!checkIfWithinWing(aLiftingSurface, yPosition)) {
        myRuntimeInfo->err << "Given YPosition is outside the wing!" << endl;
        exit(1);
    }
    size_t innerSectionID = determineWingSection(aLiftingSurface, yPosition);
    double innerThickness = geom2::measure::thickness_max(aLiftingSurface.sections.at(innerSectionID));
    double outerThickness = geom2::measure::thickness_max(aLiftingSurface.sections.at(innerSectionID + 1));
    double innerYPosition = getTrueCoordinates(aLiftingSurface, innerSectionID).yCoordinate;
    double outerYPosition = getTrueCoordinates(aLiftingSurface, innerSectionID + 1).yCoordinate;
    // law of the levers
    double localThickness = (fabs(outerYPosition - yPosition) * innerThickness + fabs(yPosition - innerYPosition) * outerThickness) / fabs(outerYPosition - innerYPosition);
    return localThickness;
}

double getLocalSweep(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition, const double relChordPosition, std::string angularUnit) {
    if (!checkIfWithinWing(aLiftingSurface, yPosition)) {
        myRuntimeInfo->err << "Given YPosition is outside the wing!" << endl;
        exit(1);
    }
    if (angularUnit != "RADIAN" && angularUnit != "DEGREE") {
        myRuntimeInfo->err << "auxFunctions::getLocalSweep given angularUnit is neither RAD or DEG, abort!" << endl;
    }
    double CGAL_ZCoordinate = getCGAL_ZCoordinate(aLiftingSurface, yPosition);
    double sweep = -1 * geom2::measure::sweep(aLiftingSurface, CGAL_ZCoordinate, relChordPosition);
    if (angularUnit == "DEGREE") {
        sweep = convertUnit(RADIAN, DEGREE, sweep);
    }
    return sweep;
}

double getAverageTtoC(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface) {
    double avgTtoC(0.0);
    std::vector<double> chordLenghts;
    std::vector<double> thicknesses;
    std::vector<double> segmentSpan;
    for (size_t sectionID(0); sectionID < aLiftingSurface.sections.size(); ++sectionID) {
        chordLenghts.push_back(aLiftingSurface.sections.at(sectionID).get_chord_length());
        thicknesses.push_back(geom2::measure::thickness_max(aLiftingSurface.sections.at(sectionID)));
        if (sectionID > 0) {
            segmentSpan.push_back(sqrt(pow((getTrueCoordinates(aLiftingSurface, sectionID).yCoordinate - getTrueCoordinates(aLiftingSurface, sectionID - 1).yCoordinate), 2)
                                    + pow((getTrueCoordinates(aLiftingSurface, sectionID).zCoordinate - getTrueCoordinates(aLiftingSurface, sectionID - 1).zCoordinate), 2)));
        }
    }
    std::vector<double>segmentavgTtoC;
    for (size_t segmentID(0); segmentID < segmentSpan.size(); ++segmentID) {
        segmentavgTtoC.push_back(((thicknesses.at(segmentID) / chordLenghts.at(segmentID)) + (thicknesses.at(segmentID + 1) / chordLenghts.at(segmentID + 1))) * 0.5);
    }
    for (size_t segmentID(0); segmentID < segmentSpan.size(); ++segmentID) {
        avgTtoC += segmentavgTtoC.at(segmentID) * segmentSpan.at(segmentID);
    }
    return avgTtoC / std::accumulate(segmentSpan.begin(), segmentSpan.end(), 0.0);
}

double getReynoldsNumber(const double& length, const double& altitude, const double& mach) {
    atmosphere myISA;
    double dynamicViscosity(myISA.getViscosity(altitude));
    double density(myISA.getDensity(altitude));
    double velocity(convertUnit(MACH, METERPERSECOND, altitude, myISA, mach));
        if (accuracyCheck(fabs(dynamicViscosity), 0.0, ACCURACY_HIGH)) {  // check if denominator is zero
        throwError(__FILE__, __func__, __LINE__, "Cannot predict Reynolds number for zero dyn. viscosity. Abort program.");
    }
    return (density * length * velocity / dynamicViscosity);
    return 0.0;
}

/* Determination of spanwise Cl from lift distribution for current y-Pos  */
double getClFromLiftDistribution(double yPos, std::vector <double> yPosVec, std::vector <double> aLiftDistr) {
    double ClTmp(NAN);
    for (size_t k(0); k < yPosVec.size() - 1; ++k) {
        if (fabs(yPos - yPosVec.at(k)) < ACCURACY_MEDIUM) {  // on a y(LL)-station
            ClTmp = aLiftDistr.at(k);
            break;
        } else if ((yPos - yPosVec.at(k)) >= -ACCURACY_MEDIUM && (yPos - yPosVec.at(k + 1)) < ACCURACY_MEDIUM) {  // between 2 y(LL)-Stations (applies to inserted Geo-Sections)
            double x((yPos - yPosVec.at(k)) / (yPosVec.at(k + 1) - yPosVec.at(k)));
            ClTmp = (1. - x) * aLiftDistr.at(k) + x * aLiftDistr.at(k + 1);
            break;
        } else if (yPos > yPosVec.back()) {  // above (right) y(LL)-Stations
            int lastID = yPosVec.size() - 1;
            double x((yPos - yPosVec.at(lastID - 1)) / (yPosVec.at(lastID) - yPosVec.at(lastID - 1)));
            ClTmp = (1. - x) * aLiftDistr.at(lastID - 1) + x * aLiftDistr.at(lastID);
            break;
        } else if (yPos < yPosVec.front()) {  // below (left) y(LL)-Stations
            double x((yPos - yPosVec.at(0)) / (yPosVec.at(1) - yPosVec.at(0)));
            ClTmp = (1. - x) * aLiftDistr.at(0) + x * aLiftDistr.at(1);
            break;
        } else {}
    }
    return ClTmp;
}

double getThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const size_t sectionID) {
    double thickness = fabs(aLiftingSurface.sections.at(sectionID).get_contour().bottom_vertex()->y() - aLiftingSurface.sections.at(sectionID).get_contour().top_vertex()->y());
    double chord = fabs(aLiftingSurface.sections.at(sectionID).get_contour().right_vertex()->x() - aLiftingSurface.sections.at(sectionID).get_contour().left_vertex()->x());
    return thickness / chord;
}

double getAverageSweep(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, double relChord) {
    double averageSweep(0.0);
    std::vector<double> sweeps;
    std::vector<double> segmentSpans;
    for (size_t sectionID(0); sectionID < aLiftingSurface.sections.size(); ++sectionID) {
        if (sectionID > 0) {
            segmentSpans.push_back(sqrt(pow((getTrueCoordinates(aLiftingSurface, sectionID).yCoordinate - getTrueCoordinates(aLiftingSurface, sectionID - 1).yCoordinate), 2)
                                    + pow((getTrueCoordinates(aLiftingSurface, sectionID).zCoordinate - getTrueCoordinates(aLiftingSurface, sectionID - 1).zCoordinate), 2)));
            double yPosition = 0.5 * (getTrueCoordinates(aLiftingSurface, sectionID).yCoordinate + getTrueCoordinates(aLiftingSurface, sectionID - 1).yCoordinate);
            sweeps.push_back(getLocalSweep(aLiftingSurface, yPosition, relChord, "DEGREE"));
            averageSweep += segmentSpans.back() * sweeps.back();
        }
    }
    return averageSweep / std::accumulate(segmentSpans.begin(), segmentSpans.end(), 0.0);
}

double getHalfSpan(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface) {
    double halfSpan = fabs(aLiftingSurface.sections.front().origin.z() - aLiftingSurface.sections.back().origin.z());
    return halfSpan;
}

double getNacelleWettedArea(const geom2::MultisectionSurface<geom2::PolygonSection>& aNacelle) {
    double max_width = geom2::measure::width_max(aNacelle);
    double max_height = geom2::measure::height_max(aNacelle);
    double length = geom2::measure::length(aNacelle);
    // should work good enough for nearly circular nacelle
    return length * PI * 0.5 * (max_width + max_height);
}

double getNeutralPointXposition(const std::vector<defaultPolarPoint>& cruisePolarPoints, const double& mean_aerodynamic_chord, const double& trim_reference_point_xCoord){
    vector<double> alpha;
    vector<double> lift_coefficient;
    vector<double> moment_coefficient;
    // Collect columns alpha, cl, and cm of polar
    for (const defaultPolarPoint& aPolarPoint : cruisePolarPoints) {
        alpha.push_back(aPolarPoint.AoA);
        lift_coefficient.push_back(aPolarPoint.CL);
        moment_coefficient.push_back(aPolarPoint.CM);
    }
    //Calculate gradient dCM/dCL (as static_margin) by qr decomposition
    double static_margin(calcRegressionCoefficientsUsingQRdecomp(lift_coefficient, moment_coefficient, 1).at(1)); // f(x) = m * x + c with vector={c, m}
    // std::cout << std::setprecision(5) << "dCM/dCL: " << static_margin << std::endl;
    //Calculate neutral point
    double neutral_point_xCoordinate(trim_reference_point_xCoord - static_margin * mean_aerodynamic_chord);
    // std::cout << "Global MAC pos: " << neutral_point_xCoordinate << std::endl;
    return neutral_point_xCoordinate;
}

std::vector<defaultPolarPoint> getTrimmedCruisePolar(const std::unordered_map<size_t, tawDefaultPolar>& theTrimmedCleanPolars, const double& initialCruiseMachNumber){
    std::vector<defaultPolarPoint> theTrimmedCruisePolar(theTrimmedCleanPolars.at(0).Points);
    for (size_t polarID(0); polarID < theTrimmedCleanPolars.size(); ++polarID) {
        if (accuracyCheck(theTrimmedCleanPolars.at(polarID).MachNumber, initialCruiseMachNumber, ACCURACY_HIGH)) {
            theTrimmedCruisePolar = theTrimmedCleanPolars.at(polarID).Points;
        }
    }
    return theTrimmedCruisePolar;
}

std::vector<defaultPolarPoint> getTrimmedTandemCruisePolar(const std::unordered_map<size_t, tandemDefaultPolar>& theTrimmedCleanPolars, const double& initialCruiseMachNumber){
    std::vector<defaultPolarPoint> theTrimmedCruisePolar(theTrimmedCleanPolars.at(0).Points);
    for (size_t polarID(0); polarID < theTrimmedCleanPolars.size(); ++polarID) {
        if (accuracyCheck(theTrimmedCleanPolars.at(polarID).MachNumber, initialCruiseMachNumber, ACCURACY_HIGH)) {
            theTrimmedCruisePolar = theTrimmedCleanPolars.at(polarID).Points;
        }
    }
    return theTrimmedCruisePolar;
}