/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CALCULATEPOLAR_SRC_METHODS_GENERALMETHODS_AUXFUNCTIONS_H_
#define CALCULATEPOLAR_SRC_METHODS_GENERALMETHODS_AUXFUNCTIONS_H_

#include <standardFiles/typedefs.h>
#include <string>
#include <vector>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <atmosphere/atmosphere.h>

// this class is mostly just there to cope with the CGAL madness of the aircraftgeometry2
class auxFunctions;
class tawDefaultPolar;
class tandemDefaultPolar;
class defaultPolarPoint;
    /** \brief method, which transforms the cursed CGAL Coordinates into aircraft coordinates
     *  \param aLiftingSurface the Lifting Surface, which must not have an additional parent
     *  \param sectionID the section for which the leading edge corrdinates get returned
     *  \return The 3D aircraft Coordinate of the leading edge at the given section
     */
    Point getTrueCoordinates(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const size_t& sectionID);

    Point get_l_mu_ref_Point(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface);

    /** \brief this is just the inversion of getTrueCoordinates
     *  \param aLiftingSurface the lifting surface which is parent for the input coordinate
     *  \param yPosition the span postion in [m] which gets transformed into CGAL
     *  \return the z-Coordinate which is used in CGAL related functions
     */
    double getCGAL_ZCoordinate(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);

    bool checkIfWithinWing(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);

    /** \brief returns the inner wing section corresponding to the given yPosition
     */
    size_t determineWingSection(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);

    double getLocalChordLength(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);

    /** \brief returns the twist in degree at a given span position
     * \return twist in [deg]
     */
    double getLocalTwist(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);

    double getLocalThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition);
    /** \brief returns the sweep true sweep instead of the mathematical angle from CGAL
     * \return sweep in deg or rad
     */
    double getLocalSweep(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const double yPosition, const double relChordPosition, std::string angularUnit);

    double getAverageTtoC(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface);
    /** \brief returns the reynolds number
     * \param length caracteristic length in [m]
     * \param altitude flight altitude in [m]
     * \param mach fligh mach number in [-]
     * \return reynolds number
     */
    double getReynoldsNumber(const double& length, const double& altitude, const double& mach);

    double getClFromLiftDistribution(double yPos, std::vector <double> yPosVec, std::vector <double> aLiftDistr);

    double getThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, const size_t sectionID);

    /** \brief returns averages sweep line over whole lifting surface
     * \return sweep in [deg]
     */
    double getAverageSweep(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface, double relChord);
    /** \brief returns the half span of a lifting surface
     * \return half span in [m]
     */
    double getHalfSpan(const geom2::MultisectionSurface<geom2::AirfoilSection>& aLiftingSurface);
    /** \brief returns the wettet area of a nacelle, simplified by the max with, max height and length
     * \return half area in [m^2]
     */
    double getNacelleWettedArea(const geom2::MultisectionSurface<geom2::PolygonSection>& aNacelle);
    /** \brief returns the aircraft's neutral point relative to MAC for the trimmed cruise polar
     * \param cruisePolarPoints const std::vector<defaultPolarPoint>&: constant reference to points of trimmed cruise polar
     * \param mean_aerodynamic_chord const double&: constant reference to the mean aerodynamic chord
     * \param trim_reference_point_xCoord const double&: constant reference to the x Position of trim reference point [m]
     * \return neutral_point_xCoordinate double: aircraft's neutral point x coordinate in global aircraft reference coordinate system [m]
     */
    double getNeutralPointXposition(const std::vector<defaultPolarPoint>& cruisePolarPoints, const double& mean_aerodynamic_chord, const double& trim_reference_point_xCoord);
    /** \brief returns the aircraft's trimmed cruise polar
     * \param theTrimmedCleanPolars const std::unordered_map<size_t, tawDefaultPolar>&: constant reference to all trimmed clean polars
     * \param initialCruiseMachNumber const double&: constant reference to the initial cruise Mach number
     * \return theTrimmedCruisePolar std::vector<defaultPolarPoint>: points of the aircraft's trimmed cruise polar [-]
     */
    std::vector<defaultPolarPoint> getTrimmedCruisePolar(const std::unordered_map<size_t, tawDefaultPolar>& theTrimmedCleanPolars, const double& initialCruiseMachNumber);
    /** \brief returns the tandem wing aircraft's trimmed cruise polar
     * \param theTrimmedCleanPolars const std::unordered_map<size_t, tandemDefaultPolar>&: constant reference to all trimmed clean polars
     * \param initialCruiseMachNumber const double&: constant reference to the initial cruise Mach number
     * \return theTrimmedCruisePolar std::vector<defaultPolarPoint>: points of the aircraft's trimmed cruise polar [-]
     */
    std::vector<defaultPolarPoint> getTrimmedTandemCruisePolar(const std::unordered_map<size_t, tandemDefaultPolar>& theTrimmedCleanPolars, const double& initialCruiseMachNumber);


#endif  // CALCULATEPOLAR_SRC_METHODS_GENERALMETHODS_AUXFUNCTIONS_H_
