/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tandemHighLiftAdaption.h"

#include <aircraftGeometry2/processing/measure.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <memory>
#include <utility>  // make_pair

#include "../polar.h"
#include "generalMethods/auxFunctions.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::vector;

/* TODO:        find consistent way to account for "special" high lift devices - or not
*/
tandemHighLiftAdaption::tandemHighLiftAdaption(const std::shared_ptr<RuntimeIO>& rtIO,  const geom2::MultisectionSurface<geom2::AirfoilSection> frontWing, const geom2::MultisectionSurface<geom2::AirfoilSection> backWing)
    :
    CLmaxTO_req(0.),
    CLmaxLDG_req(0.),
    stepWidthCL(0.),
    CDmin_inc(0.),
    kFactor_inc(0.),
    CLatCDmin_inc(0.),
    reductionDragCountsHighLift(0.),
    factorDragHighLift(0.),
    dCMtodCL_inc(0.),
    CMatCL0_inc(0.),
    dCMtodAoA(0.),
    dCLtodAoA(0.),
    CLatAoA0(0.),
    CMatAoA0(0.),
    theFrontWing(frontWing),
    theBackWing(backWing) {
}

void tandemHighLiftAdaption::runHighLiftAdaption() {
    myRuntimeInfo->out << "Run high lift calculation" << endl;
    // check if Mach number of provided polar is <= 0.2
    if (theCleanPolar.MachNumber > 0.2) {
        myRuntimeInfo->warn << "Mach Number of clean polar for high lift calculation is " << theCleanPolar.MachNumber << ", wich is larger than 0.2" << endl;
    }
    this->setHighLiftDevices();
    this->processHighLiftDevices();
    this->setDeltaCLMaxDevices();
}

void tandemHighLiftAdaption::setHighLiftDevices() {
    // sort out LE and TE high lift devices
    for (size_t deviceID(0); deviceID < front_wingDevices.size(); ++deviceID) {
        if (front_wingDevices.at(deviceID).name == "droop_aileron"  ||
        front_wingDevices.at(deviceID).name == "flap"               ||
        front_wingDevices.at(deviceID).name == "slotted_flap"       ||
        front_wingDevices.at(deviceID).name == "flap_fowler"        ||
        front_wingDevices.at(deviceID).name == "double_fowler"      ||
        front_wingDevices.at(deviceID).name == "triple_fowler"      ||
        front_wingDevices.at(deviceID).name == "ADHF"                 ) {
        front_trailingEdgeHighLiftDevices.push_back(front_wingDevices.at(deviceID));
        }
        if (front_wingDevices.at(deviceID).name == "slat"           ||
        front_wingDevices.at(deviceID).name == "krueger"            ||
        front_wingDevices.at(deviceID).name == "droop_nose"         ||
        front_wingDevices.at(deviceID).name == "morphing_droop_nose" ) {
        front_leadingEdgeHighLiftDevices.push_back(front_wingDevices.at(deviceID));
        }
    }
    for (size_t deviceID(0); deviceID < back_wingDevices.size(); ++deviceID) {
        if (back_wingDevices.at(deviceID).name == "droop_aileron"  ||
        back_wingDevices.at(deviceID).name == "flap"               ||
        back_wingDevices.at(deviceID).name == "slotted_flap"       ||
        back_wingDevices.at(deviceID).name == "flap_fowler"        ||
        back_wingDevices.at(deviceID).name == "double_fowler"      ||
        back_wingDevices.at(deviceID).name == "triple_fowler"      ||
        back_wingDevices.at(deviceID).name == "ADHF"                 ) {
        back_trailingEdgeHighLiftDevices.push_back(back_wingDevices.at(deviceID));
        }
        if (back_wingDevices.at(deviceID).name == "slat"           ||
        back_wingDevices.at(deviceID).name == "krueger"            ||
        back_wingDevices.at(deviceID).name == "droop_nose"         ||
        back_wingDevices.at(deviceID).name == "morphing_droop_nose" ) {
        back_leadingEdgeHighLiftDevices.push_back(back_wingDevices.at(deviceID));
        }
    }
}

void tandemHighLiftAdaption::processHighLiftDevices() {   
    for (size_t deviceID(0); deviceID < front_leadingEdgeHighLiftDevices.size(); ++deviceID) {
        front_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs = getHalfSpan(theFrontWing) * front_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_rel;
        front_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs = getHalfSpan(theFrontWing) * front_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_rel;
        front_leadingEdgeHighLiftDevices.at(deviceID).span = fabs(front_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs - front_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs);
        front_leadingEdgeHighLiftDevices.at(deviceID).c_rel_avg = 0.25 * (front_leadingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition
                                                                + front_leadingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition
                                                                + front_leadingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition
                                                                + front_leadingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition);
        double delta_x_rel_HL = 0.02;  // distance hinge line (apparently from Raymer..)
        front_leadingEdgeHighLiftDevices.at(deviceID).x_rel_hinge = front_leadingEdgeHighLiftDevices.at(deviceID).c_rel_avg - delta_x_rel_HL;
        front_leadingEdgeHighLiftDevices.at(deviceID).phiHL = getLocalSweep(theFrontWing,
                                                            0.5 * (front_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs + front_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs),
                                                            front_leadingEdgeHighLiftDevices.at(deviceID).x_rel_hinge, "DEGREE");
        double chordInboard = getLocalChordLength(theFrontWing, front_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs)
                                    * fabs(front_leadingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition - front_leadingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition);
        double chordOutboard = getLocalChordLength(theFrontWing, front_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs)
                                    * fabs(front_leadingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition - front_leadingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition);
        // area is calculated for both devices, because symmetry
        front_leadingEdgeHighLiftDevices.at(deviceID).coveredA = front_leadingEdgeHighLiftDevices.at(deviceID).span * (chordInboard + chordOutboard);
    }
    for (size_t deviceID(0); deviceID < front_trailingEdgeHighLiftDevices.size(); ++deviceID) {
        front_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs = getHalfSpan(theFrontWing) * front_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_rel;
        front_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs = getHalfSpan(theFrontWing) * front_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_rel;
        front_trailingEdgeHighLiftDevices.at(deviceID).span = fabs(front_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs - front_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs);
        front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg = 0.25 * (front_trailingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition
                                                                + front_trailingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition
                                                                + front_trailingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition
                                                                + front_trailingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition);
        double delta_x_rel_HL = 0.02;  // distance hinge line (apparently from Raymer..)
        front_trailingEdgeHighLiftDevices.at(deviceID).x_rel_hinge = 1. - front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg + delta_x_rel_HL;
        front_trailingEdgeHighLiftDevices.at(deviceID).phiHL = getLocalSweep(theFrontWing,
                                                            0.5 * (front_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs + front_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs),
                                                            front_trailingEdgeHighLiftDevices.at(deviceID).x_rel_hinge, "DEGREE");
        double chordInboard = getLocalChordLength(theFrontWing, front_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs)
                                    * fabs(front_trailingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition - front_trailingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition);
        double chordOutboard = getLocalChordLength(theFrontWing, front_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs)
                                    * fabs(front_trailingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition - front_trailingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition);
        // area is calculated for both devices, because symmetry
        front_trailingEdgeHighLiftDevices.at(deviceID).coveredA = front_trailingEdgeHighLiftDevices.at(deviceID).span * (chordInboard + chordOutboard);
    }
    for (size_t deviceID(0); deviceID < back_leadingEdgeHighLiftDevices.size(); ++deviceID) {
        back_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs = getHalfSpan(theBackWing) * back_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_rel;
        back_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs = getHalfSpan(theBackWing) * back_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_rel;
        back_leadingEdgeHighLiftDevices.at(deviceID).span = fabs(back_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs - back_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs);
        back_leadingEdgeHighLiftDevices.at(deviceID).c_rel_avg = 0.25 * (back_leadingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition
                                                                + back_leadingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition
                                                                + back_leadingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition
                                                                + back_leadingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition);
        double delta_x_rel_HL = 0.02;  // distance hinge line (apparently from Raymer..)
        back_leadingEdgeHighLiftDevices.at(deviceID).x_rel_hinge = back_leadingEdgeHighLiftDevices.at(deviceID).c_rel_avg - delta_x_rel_HL;
        back_leadingEdgeHighLiftDevices.at(deviceID).phiHL = getLocalSweep(theBackWing,
                                                            0.5 * (back_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs + back_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs),
                                                            back_leadingEdgeHighLiftDevices.at(deviceID).x_rel_hinge, "DEGREE");
        double chordInboard = getLocalChordLength(theBackWing, back_leadingEdgeHighLiftDevices.at(deviceID).innerSpan_abs)
                                    * fabs(back_leadingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition - back_leadingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition);
        double chordOutboard = getLocalChordLength(theBackWing, back_leadingEdgeHighLiftDevices.at(deviceID).outerSpan_abs)
                                    * fabs(back_leadingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition - back_leadingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition);
        // area is calculated for both devices, because symmetry
        back_leadingEdgeHighLiftDevices.at(deviceID).coveredA = back_leadingEdgeHighLiftDevices.at(deviceID).span * (chordInboard + chordOutboard);
    }
    for (size_t deviceID(0); deviceID < back_trailingEdgeHighLiftDevices.size(); ++deviceID) {
        back_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs = getHalfSpan(theBackWing) * back_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_rel;
        back_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs = getHalfSpan(theBackWing) * back_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_rel;
        back_trailingEdgeHighLiftDevices.at(deviceID).span = fabs(back_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs - back_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs);
        back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg = 0.25 * (back_trailingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition
                                                                + back_trailingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition
                                                                + back_trailingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition
                                                                + back_trailingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition);
        double delta_x_rel_HL = 0.02;  // distance hinge line (apparently from Raymer..)
        back_trailingEdgeHighLiftDevices.at(deviceID).x_rel_hinge = 1. - back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg + delta_x_rel_HL;
        back_trailingEdgeHighLiftDevices.at(deviceID).phiHL = getLocalSweep(theBackWing,
                                                            0.5 * (back_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs + back_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs),
                                                            back_trailingEdgeHighLiftDevices.at(deviceID).x_rel_hinge, "DEGREE");
        double chordInboard = getLocalChordLength(theBackWing, back_trailingEdgeHighLiftDevices.at(deviceID).innerSpan_abs)
                                    * fabs(back_trailingEdgeHighLiftDevices.at(deviceID).innerRelAftChordPosition - back_trailingEdgeHighLiftDevices.at(deviceID).innerRelFwdChordPosition);
        double chordOutboard = getLocalChordLength(theBackWing, back_trailingEdgeHighLiftDevices.at(deviceID).outerSpan_abs)
                                    * fabs(back_trailingEdgeHighLiftDevices.at(deviceID).outerRelAftChordPosition - back_trailingEdgeHighLiftDevices.at(deviceID).outerRelFwdChordPosition);
        // area is calculated for both devices, because symmetry
        back_trailingEdgeHighLiftDevices.at(deviceID).coveredA = back_trailingEdgeHighLiftDevices.at(deviceID).span * (chordInboard + chordOutboard);
    }
}

void tandemHighLiftAdaption::setDeltaCLMaxDevices() { //TODO kayra: dimensionless yaparken q degisimi ve downwash dikkate alamadik
    for (size_t deviceID(0); deviceID < front_leadingEdgeHighLiftDevices.size(); ++deviceID) {
        // Determination of deltaCLmax
        // Method [Ray18], p.277
        double localDeltaLift(0.0);
        if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "slat") {
            localDeltaLift = 0.4 * 1.05;
        } else if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "krueger") {
            localDeltaLift = 0.3;
        } else if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "droop_nose") {
            localDeltaLift = 0.3;
        } else if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "morphing_droop_nose") {
            /** \todo [T777] (https://unicado.ilr.rwth-aachen.de/T777) Add aero-performance for morphing droop nose
             * \note What is meant with aero-performance here? (This task might already be resolved by the localDeltaLift = 0.1) */            
            localDeltaLift = 0.1;
        } else if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "none") {
            localDeltaLift = 0.0;
        } else if (front_leadingEdgeHighLiftDevices.at(deviceID).name == "special") {
            localDeltaLift = 0.0;
        } else {
            myRuntimeInfo->err << "LE type not known. deltaLocalLift cannot be determined!" << endl;
            myRuntimeInfo->err << front_leadingEdgeHighLiftDevices.at(deviceID).name << endl;
            exit(1);
        }
        front_leadingEdgeHighLiftDevices.at(deviceID).deltaMaxCL = localDeltaLift * front_leadingEdgeHighLiftDevices.at(deviceID).coveredA
                                                            / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing))
                                                            * cos(convertUnit(DEGREE, RADIAN, front_leadingEdgeHighLiftDevices.at(deviceID).phiHL));
    }
    for (size_t deviceID(0); deviceID < front_trailingEdgeHighLiftDevices.size(); ++deviceID) {
        // Determination of deltaCLmax
        // Method [Ray18], p.277
        double localDeltaLift(0.0);
        double chordFactor(0.);  // [Tor13] Fig. G-9
        if (front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg <= 0.2) {
            // Assumption from diagram: linear between 0.90 at 0.1 and 1.0 at 0.2
            //  (instead of ca. 0.8 -> decrease not as much as specified by TB; in general, otherwise the short rel. chord lengths inboard would be too much disadvantaged)
            chordFactor = 1. + (0.90 - 1.) / (0.1 - 0.2) * (front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg - 0.2);
        } else {
            // Assumption from diagram: linear between 1.0 at 0.2 and 1.05 at 0.3
            chordFactor = 1. + (1.05 - 1.) / (0.3 - 0.2) * (front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg - 0.2);
        }
        if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "aileron") {
            localDeltaLift = 0.0;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "droop_aileron") {
            localDeltaLift = 0.4;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "flap") {
            localDeltaLift = 0.9;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "slotted_flap") {
            localDeltaLift = 1.3;  // [Ray18] table 12.2
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "flap_fowler") {
            double chord_ext_rel2flapChord(0.5);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.3 * chordFactor * (1. + chord_ext_rel2flapChord * front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "ADHF") {
            localDeltaLift = 1.4;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "double_fowler") {
            double chord_ext_rel2flapChord(0.6);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.6 * chordFactor * (1. + chord_ext_rel2flapChord * front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "triple_fowler") {
            double chord_ext_rel2flapChord(0.7);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.9 * chordFactor * (1. + chord_ext_rel2flapChord * front_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "morphing trailing edge") {
            /** \todo [T780] (https://unicado.ilr.rwth-aachen.de/T780) Insert aero-performance for morphing trailing edge
            */
            localDeltaLift = 0.0;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "none") {
            localDeltaLift = 0.0;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "special") {
            localDeltaLift = 0.0;
        } else if (front_trailingEdgeHighLiftDevices.at(deviceID).name == "spoiler_ground" || front_trailingEdgeHighLiftDevices.at(deviceID).name == "spoiler_air") {
            localDeltaLift = 0.0;
        } else {
            myRuntimeInfo->err << "TE type not known. deltaLocalLift cannot be determined!" << endl;
            myRuntimeInfo->err << front_trailingEdgeHighLiftDevices.at(deviceID).name << endl;
            exit(1);
        }
        front_trailingEdgeHighLiftDevices.at(deviceID).deltaMaxCL = localDeltaLift * front_trailingEdgeHighLiftDevices.at(deviceID).coveredA
                                                                / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing))
                                                                * cos(convertUnit(DEGREE, RADIAN, front_trailingEdgeHighLiftDevices.at(deviceID).phiHL));
    }
    for (size_t deviceID(0); deviceID < back_leadingEdgeHighLiftDevices.size(); ++deviceID) {  //TODO kayra: ön kanat daha önce stalla girerse hesap yanlis olmaz mi: yoksa ll gerekir tekrar
        // Determination of deltaCLmax
        // Method [Ray18], p.277
        double localDeltaLift(0.0);
        if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "slat") {
            localDeltaLift = 0.4 * 1.05;
        } else if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "krueger") {
            localDeltaLift = 0.3;
        } else if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "droop_nose") {
            localDeltaLift = 0.3;
        } else if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "morphing_droop_nose") {
            /** \todo [T777] (https://unicado.ilr.rwth-aachen.de/T777) Add aero-performance for morphing droop nose
             * \note What is meant with aero-performance here? (This task might already be resolved by the localDeltaLift = 0.1) */            
            localDeltaLift = 0.1;
        } else if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "none") {
            localDeltaLift = 0.0;
        } else if (back_leadingEdgeHighLiftDevices.at(deviceID).name == "special") {
            localDeltaLift = 0.0;
        } else {
            myRuntimeInfo->err << "LE type not known. deltaLocalLift cannot be determined!" << endl;
            myRuntimeInfo->err << back_leadingEdgeHighLiftDevices.at(deviceID).name << endl;
            exit(1);
        }
        back_leadingEdgeHighLiftDevices.at(deviceID).deltaMaxCL = localDeltaLift * back_leadingEdgeHighLiftDevices.at(deviceID).coveredA
                                                            / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing))
                                                            * cos(convertUnit(DEGREE, RADIAN, back_leadingEdgeHighLiftDevices.at(deviceID).phiHL));
    }
    for (size_t deviceID(0); deviceID < back_trailingEdgeHighLiftDevices.size(); ++deviceID) {
        // Determination of deltaCLmax
        // Method [Ray18], p.277
        double localDeltaLift(0.0);
        double chordFactor(0.);  // [Tor13] Fig. G-9
        if (back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg <= 0.2) {
            // Assumption from diagram: linear between 0.90 at 0.1 and 1.0 at 0.2
            //  (instead of ca. 0.8 -> decrease not as much as specified by TB; in general, otherwise the short rel. chord lengths inboard would be too much disadvantaged)
            chordFactor = 1. + (0.90 - 1.) / (0.1 - 0.2) * (back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg - 0.2);
        } else {
            // Assumption from diagram: linear between 1.0 at 0.2 and 1.05 at 0.3
            chordFactor = 1. + (1.05 - 1.) / (0.3 - 0.2) * (back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg - 0.2);
        }
        if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "aileron") {
            localDeltaLift = 0.0; 
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "droop_aileron") {
            localDeltaLift = 0.4;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "flap") {
            localDeltaLift = 0.9;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "slotted_flap") {
            localDeltaLift = 1.3;  // [Ray18] table 12.2
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "flap_fowler") {
            double chord_ext_rel2flapChord(0.5);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.3 * chordFactor * (1. + chord_ext_rel2flapChord * back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "ADHF") {
            localDeltaLift = 1.4;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "double_fowler") {
            double chord_ext_rel2flapChord(0.6);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.6 * chordFactor * (1. + chord_ext_rel2flapChord * back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "triple_fowler") {
            double chord_ext_rel2flapChord(0.7);  // [Tor13] Fig. G.-7
            localDeltaLift = 1.9 * chordFactor * (1. + chord_ext_rel2flapChord * back_trailingEdgeHighLiftDevices.at(deviceID).c_rel_avg);
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "morphing trailing edge") {
            /** \todo [T780] (https://unicado.ilr.rwth-aachen.de/T780) Insert aero-performance for morphing trailing edge
            */
            localDeltaLift = 0.0;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "none") {
            localDeltaLift = 0.0;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "special") {
            localDeltaLift = 0.0;
        } else if (back_trailingEdgeHighLiftDevices.at(deviceID).name == "spoiler_ground" || back_trailingEdgeHighLiftDevices.at(deviceID).name == "spoiler_air") {
            localDeltaLift = 0.0;
        } else {
            myRuntimeInfo->err << "TE type not known. deltaLocalLift cannot be determined!" << endl;
            myRuntimeInfo->err << back_trailingEdgeHighLiftDevices.at(deviceID).name << endl;
            exit(1);
        }
        back_trailingEdgeHighLiftDevices.at(deviceID).deltaMaxCL = localDeltaLift * back_trailingEdgeHighLiftDevices.at(deviceID).coveredA
                                                                / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing))
                                                                * cos(convertUnit(DEGREE, RADIAN, back_trailingEdgeHighLiftDevices.at(deviceID).phiHL));
    }
}

Polar tandemHighLiftAdaption::calculateHighLiftPolar(std::string config, const double CoG_X_position, const double currentIBackWingAngle) {
    this->theHighliftAdaption = this->setRaymerHighLiftAdaptionValues();
    Polar aHighLiftPolar;
    aHighLiftPolar.extrapolationMargin = theCleanPolar.extrapolationMargin;
    aHighLiftPolar.allowGridChange = theCleanPolar.allowGridChange;
    aHighLiftPolar.MachNumber = theCleanPolar.MachNumber;
    aHighLiftPolar.altitude = theCleanPolar.altitude;
    aHighLiftPolar.ReynoldsNumber = theCleanPolar.ReynoldsNumber;
    aHighLiftPolar.Config = config;
    double CLmax = getCLMax(theCleanPolar);
    double deltaMaxCL_front = 0.0;
    double deltaMaxCL_back = 0.0;
    // Set high-lift coefficients
    double factorLE = this->theHighliftAdaption.at(config).factorLE;
    double factorTE = this->theHighliftAdaption.at(config).factorTE;
    double deltaCDmin = this->theHighliftAdaption.at(config).deltaCDmin;
    double deltaCDgear = this->theHighliftAdaption.at(config).deltaCDgear;
    double deltaCLatCDmin = this->theHighliftAdaption.at(config).deltaCLatCDmin;
    // Determination of CLmax
    bool specialDevice(false);
    for (size_t i(0); i < front_leadingEdgeHighLiftDevices.size(); ++i) {
        deltaMaxCL_front += factorLE * front_leadingEdgeHighLiftDevices.at(i).deltaMaxCL;
        if (front_leadingEdgeHighLiftDevices.at(i).name == "special") {
            specialDevice = true;
        }
    }
    for (size_t i(0); i < back_leadingEdgeHighLiftDevices.size(); ++i) {
        deltaMaxCL_back += factorLE * back_leadingEdgeHighLiftDevices.at(i).deltaMaxCL;
        if (back_leadingEdgeHighLiftDevices.at(i).name == "special") {
            specialDevice = true;
        }
    }
    for (size_t i(0); i < front_trailingEdgeHighLiftDevices.size(); ++i) {
        deltaMaxCL_front += factorTE * front_trailingEdgeHighLiftDevices.at(i).deltaMaxCL;
        if (front_trailingEdgeHighLiftDevices.at(i).name == "special") {
            specialDevice = true;
        }
    }
    for (size_t i(0); i < back_trailingEdgeHighLiftDevices.size(); ++i) {
        deltaMaxCL_back += factorTE * back_trailingEdgeHighLiftDevices.at(i).deltaMaxCL;
        if (back_trailingEdgeHighLiftDevices.at(i).name == "special") {
            specialDevice = true;
        }
    }
    CLmax += deltaMaxCL_front;
    CLmax += deltaMaxCL_back;
    if (specialDevice == true) {
        if (CLmaxTO_req <= 0.5) {
            myRuntimeInfo->err << "Required CL_max_TO=" << CLmaxTO_req << " not useful in combination with HighLift-Devices of type 'Special'. Abort program!" << endl;
            exit(1);
        }
        if (CLmaxLDG_req <= 0.5) {
            myRuntimeInfo->err << "Required CL_max_LDG=" << CLmaxLDG_req << " not useful in combination with HighLift-Devices of type 'Special'. Abort program!" << endl;
            exit(1);
        }
        if (config == "takeoff") {
            CLmax = CLmaxTO_req;
        } else if (config == "takeoff_landing_gear_retracted") {
            CLmax = CLmaxTO_req;
        } else if (config == "climb") {
            CLmax = CLmaxTO_req;
        } else if (config == "landing") {
            CLmax = CLmaxLDG_req;
        } else if (config == "approach") {
            CLmax = CLmaxTO_req;
        } else if (config == "approach_landing_gear_out") {
            CLmax = CLmaxLDG_req;
        }
    }
    // Check if requirements regarding CLmax are met
    // Take-off
    if (config == "takeoff")
        myRuntimeInfo->out << "CL_max (t/o) = " << CLmax << endl;
    if (config == "takeoff" && CLmax < CLmaxTO_req) {
        myRuntimeInfo->warn << "CLmax,takeoff cannot be achieved by this HighLift system!" << endl;
        myRuntimeInfo->warn << "   -> CLmax,takeoff_required = " << CLmaxTO_req << endl;
        myRuntimeInfo->warn << "   -> CLmax,takeoff_achieved = " << CLmax << endl;
    }
    // landing
    if (config == "landing")
        myRuntimeInfo->out << "CL_max (ldg) = " << CLmax << endl;
    if (config == "landing" && CLmax < CLmaxLDG_req) {
        myRuntimeInfo->warn << "CLmax,landing cannot be achieved by this HighLift system!!" << endl;
        myRuntimeInfo->warn << "   -> CLmax,landing_required = " << CLmaxLDG_req << endl;
        myRuntimeInfo->warn << "   -> CLmax,landing_achieved = " << CLmax << endl;
    }

    // Determination of polars of CD = f(CL)
    // Determine angle of attack of high-lift polars
    // Determine vertical displacement of HighLift polars for AOA determination
    // Determine Cl-alpha ; Assumption: Cl-alpha = const.
    double CLalpha = (theCleanPolar.PolarPoints.back().CL - theCleanPolar.PolarPoints.front().CL) / (theCleanPolar.PolarPoints.back().AoA -
                     theCleanPolar.PolarPoints.front().AoA);
    // Leading Edge
    // Summation of wing area or span influenced by LE
    double totalDeltaMaxCLLE = 0.0;         
    for (size_t j = 0; j < front_leadingEdgeHighLiftDevices.size(); j++) {
        // Summing up DeltaMaxCL gain by LEs - already takes into account that the LE may be only proportionally distributed over the span.
        totalDeltaMaxCLLE += front_leadingEdgeHighLiftDevices.at(j).deltaMaxCL;
    }
    for (size_t j = 0; j < back_leadingEdgeHighLiftDevices.size(); j++) {
        // Summing up DeltaMaxCL gain by LEs - already takes into account that the LE may be only proportionally distributed over the span.
        totalDeltaMaxCLLE += back_leadingEdgeHighLiftDevices.at(j).deltaMaxCL;
    }
    // Determine Alpha Max increase by Leading Edge Devices; Assumption: Clalpha const.
    double deltaAOAMaxLE = totalDeltaMaxCLLE / CLalpha;
    // Trailing Edge
    // Assumption: -6 deg shift of ALpha Max by trailing edge device at full deflection
    // Summation of the wing area or span influenced by TE
    double totalCoveredATE = 0.0;
    for (size_t j = 0; j < front_trailingEdgeHighLiftDevices.size(); j++) {
        totalCoveredATE += front_trailingEdgeHighLiftDevices.at(j).coveredA / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing)) * cos(front_trailingEdgeHighLiftDevices.at(j).phiHL);
    }
    for (size_t j = 0; j < back_trailingEdgeHighLiftDevices.size(); j++) {
        totalCoveredATE += back_trailingEdgeHighLiftDevices.at(j).coveredA / (geom2::measure::reference_area(theFrontWing) + geom2::measure::reference_area(theBackWing)) * cos(back_trailingEdgeHighLiftDevices.at(j).phiHL);
    }
    double deltaAOAMaxTE = -6.0 * factorTE * totalCoveredATE;
    // Determine AlphaMax of Clean Polar
    double alphaMaxClean = theCleanPolar.PolarPoints.back().AoA;  // Determine maximum angle of attack in clean polar curve
    double alphaMaxHL = alphaMaxClean + deltaAOAMaxTE + deltaAOAMaxLE;  // With the above assumption, determine maximum angle of attack in the high-lift polar curve
    double CL_HL_AlphaMaxClean = 0.0;
    double deltaCLHLVerticalShift  = 0.0;
    if (alphaMaxClean >= alphaMaxHL) {
        // determine CL of Clean Polars at  AOA max of High-Lift Polars
        double CLCleanAOAHLMax(theCleanPolar.PolarPoints.front().CL);
        for (size_t i = 0; i < theCleanPolar.PolarPoints.size(); i++) {
            if (theCleanPolar.PolarPoints.front().AoA <= alphaMaxHL && theCleanPolar.PolarPoints.at(i).AoA >= alphaMaxHL) {
                CLCleanAOAHLMax = theCleanPolar.PolarPoints.at(i).CL;
                break;
            }
        }
        // Determine vertical displacement of CleanPolars - here: determination using Alpha = 0
        deltaCLHLVerticalShift =  CLmax - CLCleanAOAHLMax;
        CL_HL_AlphaMaxClean = 999.;  // set high value, so if condition is never True
    } else {
        CL_HL_AlphaMaxClean = CLmax - CLalpha * (alphaMaxHL - alphaMaxClean);
        deltaCLHLVerticalShift = CL_HL_AlphaMaxClean - theCleanPolar.PolarPoints.back().CL;
    }
    double acc(10e-4);
    for (double i = -0.5; i <= CLmax + stepWidthCL + acc; i = i + stepWidthCL) {
        aHighLiftPolar.PolarPoints.push_back(PolarPoint());
        double AOA;
        if (i <= CL_HL_AlphaMaxClean) {
            AOA = getAlphaAtCL(theCleanPolar, i - deltaCLHLVerticalShift);
        } else {
            // If in the area above Alphamax of Clean Polars
            AOA = (i - CL_HL_AlphaMaxClean) / CLalpha + alphaMaxClean;
        }
        aHighLiftPolar.PolarPoints.back().AoA = AOA;
        aHighLiftPolar.PolarPoints.back().CL = Rounding(i, 4);
        aHighLiftPolar.PolarPoints.back().CD = CDmin_inc + deltaCDmin + deltaCDgear
                + kFactor_inc * pow((i - CLatCDmin_inc + deltaCLatCDmin), 2);
        // drag correction
        aHighLiftPolar.PolarPoints.back().CD = (aHighLiftPolar.PolarPoints.back().CD + reductionDragCountsHighLift * 0.0001)
                                                                * factorDragHighLift;
        double deltaHighLift;
        if (true) {  // function hack until todo is resolved
            deltaHighLift = 0.3;  //TODO kayra: suan gecici olarak sadece bunu kullaniyor ama tandem icin kesin yanlis sonuc.
        } else {
            /** \todo [T778] (https://unicado.ilr.rwth-aachen.de/T778) Check deltaHighLift calculation against source
            */
            deltaHighLift = getDeltaCMHighLift(CoG_X_position, CLmax, aHighLiftPolar.MachNumber, getCLMax(theCleanPolar), CLmax);
        }
        aHighLiftPolar.PolarPoints.back().CM = (dCMtodCL_inc * aHighLiftPolar.PolarPoints.back().CL) + CMatCL0_inc
                                                            + 0.5 * (factorLE + factorTE) * deltaHighLift;
        aHighLiftPolar.PolarPoints.back().iStabPolar = currentIBackWingAngle;
    }
    return aHighLiftPolar;
}

double tandemHighLiftAdaption::getDeltaCMHighLift(const double CoG_X_Position, double Clift, double Mach, double CLmaxClean, double CLmaxHighLift) {
    // Correction of influence of high-lift devices on CM(CL) according to [Ros18] chapter: 8.2.6, equation 8.88
    double front_area = geom2::measure::reference_area(theFrontWing);
    double back_area = geom2::measure::reference_area(theBackWing);
    double total_area = front_area + back_area;
    double MAC_front = geom2::measure::mean_aerodynamic_chord(theFrontWing);
    double MAC_back = geom2::measure::mean_aerodynamic_chord(theBackWing);
    double DeltaCMhighlift(0.0);
    double deltaCM_LE_front(0.0);
    double deltaCM_LE_back(0.0);
    double deltaCM_TE_front(0.0);
    double deltaCM_TE_back(0.0);
    double deltaCM_H(0.0);
    vector<double> CLalphaH;
    double KA(0.50);
    double KP(0.33);
    double CliftClean(0.0);      //TODO kayra: her toplama da alan carpimi var bunlari yapmasak da ayni seye cikiyor sanirim
    double CmomClean(0.0);      //TODO kayra: bu lift farkina da ön kanat arka kanat gibi bakabilirsin formülde.
    CliftClean = (Clift - (CLmaxHighLift - CLmaxClean));  //TODO kayra: these values are nondimensionalized with Stot. In calculations of Roskam:Smainwing
    CmomClean = dCMtodAoA / dCLtodAoA *                   //TODO kayra: furthermore overall moment coefficients of aircraft have reference length front MAC
                (CliftClean - CLatAoA0) + CMatAoA0;       // TODO kayra: we scale these values so we can use Roskam formulas properly.
    // [Ros18] equation (8.74)
    deltaCM_TE_front = 0.1 * ((CoG_X_Position - get_l_mu_ref_Point(theFrontWing).xCoordinate) / MAC_front - 0.25) * Clift * (total_area/front_area)
                 + KA * (geom2::measure::aspect_ratio(theFrontWing) / 1.5) * (CLmaxHighLift - CLmaxClean) * (total_area/front_area) * tan(getAverageSweep(theFrontWing, 0.25))
                 + KP * ((-0.3) * (CLmaxHighLift - CLmaxClean) * (total_area/front_area) * pow(1.2, 2.))
                 - KP * (0.25 * CliftClean * (total_area/front_area) * (pow(1.2, 2.) - 1.2))
                 + KP * CmomClean * (total_area/front_area) * (pow(1.2, 2.) - 1.);
    // [Ros18]equation (8.75)
    deltaCM_LE_front = (-0.002 * 1.15 + 0.1 * 1.15 * MAC_front * Clift * (total_area/front_area)) * 0.8 * 25. * PI / 180.
                 + (CmomClean * (total_area/front_area) * (pow(1.2, 2.) - 1.) + 0.75 * CliftClean * (total_area/front_area) * 1.2 * 0.2);
    // [Ros18] equation (8.74)
    deltaCM_TE_back = 0.1 * ((CoG_X_Position - get_l_mu_ref_Point(theBackWing).xCoordinate) / MAC_back - 0.25) * Clift * (total_area/back_area)
                 + KA * (geom2::measure::aspect_ratio(theBackWing) / 1.5) * (CLmaxHighLift - CLmaxClean) * (total_area/back_area) * tan(getAverageSweep(theBackWing, 0.25))
                 + KP * ((-0.3) * (CLmaxHighLift - CLmaxClean) * (total_area/back_area) * pow(1.2, 2.))
                 - KP * (0.25 * CliftClean * (total_area/back_area) * (pow(1.2, 2.) - 1.2))
                 + KP * CmomClean * (MAC_front/MAC_back) * (total_area/back_area) * (pow(1.2, 2.) - 1.);
    // [Ros18]equation (8.75)
    deltaCM_LE_back = (-0.002 * 1.15 + 0.1 * 1.15 * MAC_back * Clift * (total_area/back_area)) * 0.8 * 25. * PI / 180.
                 + (CmomClean * (MAC_front/MAC_back) * (total_area/back_area) * (pow(1.2, 2.) - 1.) + 0.75 * CliftClean * (total_area/back_area) * 1.2 * 0.2);
    // [Ros18] equation (8.88)
    
        // [Ros18] equation (8.22)
        CLalphaH.push_back((2 * PI * geom2::measure::aspect_ratio(theBackWing))
                           /
                           (2 + sqrt(geom2::measure::aspect_ratio(theBackWing) * (sqrt(1. - pow(Mach, 2.)) + tan(getAverageSweep(theBackWing, 0.25))) + 4.)));
        deltaCM_H = CLalphaH.back() * (geom2::measure::reference_area(theBackWing) / geom2::measure::reference_area(theFrontWing))
                    * (get_l_mu_ref_Point(theFrontWing).xCoordinate + 0.25 * MAC_back - theFrontWing.origin.x())
                    / MAC_front * 2. * PI / 180.;
    

    // [Ros18] equation (8.88)
    DeltaCMhighlift = deltaCM_LE_front * (front_area/total_area) + deltaCM_TE_front * (front_area/total_area) + deltaCM_LE_back * (MAC_back/MAC_front) * (back_area/total_area) + deltaCM_TE_back * (MAC_back/MAC_front) * (back_area/total_area) + deltaCM_H * (front_area/total_area);
    myRuntimeInfo->out << "DeltaCMhighlift_method: " << DeltaCMhighlift << endl;
    return DeltaCMhighlift;
}


auto tandemHighLiftAdaption::setRaymerHighLiftAdaptionValues() -> std::unordered_map<std::string, highLiftAdaption> {
    std::unordered_map<std::string, highLiftAdaption> theRaymerHighLiftAdaption;

    double front_area = geom2::measure::reference_area(theFrontWing);
    double back_area = geom2::measure::reference_area(theBackWing);
    double total_area = front_area + back_area;

    /* Commentary Kristof regarding. factorLE and factor TE:
        Raymer: for takeoff flap settings, lift increments of about 60-80% (ofLDG-values) should be used
        before: proportion from Howe. Tab. 6.1 transferred
        Raymer's shares are more in line with my experience rather than simply half of it
        Shares for other configurations set to the best of our knowledge
    */
    theRaymerHighLiftAdaption["takeoff"].factorLE = 0.7;
    theRaymerHighLiftAdaption["takeoff"].factorTE = 0.7;
    theRaymerHighLiftAdaption["takeoff_landing_gear_retracted"].factorLE = 0.7;
    theRaymerHighLiftAdaption["takeoff_landing_gear_retracted"].factorTE = 0.7;
    theRaymerHighLiftAdaption["climb"].factorLE = 0.6;
    theRaymerHighLiftAdaption["climb"].factorTE = 0.6;
    theRaymerHighLiftAdaption["approach"].factorLE = 0.9;
    theRaymerHighLiftAdaption["approach"].factorTE = 0.9;
    theRaymerHighLiftAdaption["approach_landing_gear_out"].factorLE = 0.8;
    theRaymerHighLiftAdaption["approach_landing_gear_out"].factorTE = 0.8;
    theRaymerHighLiftAdaption["landing"].factorLE = 1.0;
    theRaymerHighLiftAdaption["landing"].factorTE = 1.0;

    // Method by HOWE
    theRaymerHighLiftAdaption["takeoff"].deltaCDmin = ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area) 
                                                    + ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["takeoff"].deltaCDgear = 0.03;        //TODO kayra: buna nasil scale yapicaz ttoal aircraft drag scaled with total area
    theRaymerHighLiftAdaption["takeoff"].deltaCLatCDmin = -0.2;
    theRaymerHighLiftAdaption["takeoff_landing_gear_retracted"].deltaCDmin = ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area)
                                                                            + ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["takeoff_landing_gear_retracted"].deltaCDgear = 0.00;
    theRaymerHighLiftAdaption["takeoff_landing_gear_retracted"].deltaCLatCDmin = -0.2;
    theRaymerHighLiftAdaption["climb"].deltaCDmin = ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area)
                                                    + ((0.03 * 1.0 - 0.004) / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["climb"].deltaCDgear = 0.00;
    theRaymerHighLiftAdaption["climb"].deltaCLatCDmin = -0.1;
    theRaymerHighLiftAdaption["approach"].deltaCDmin = (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area)
                                                        + (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["approach"].deltaCDgear = 0.00;
    theRaymerHighLiftAdaption["approach"].deltaCLatCDmin = -0.3;
    theRaymerHighLiftAdaption["approach_landing_gear_out"].deltaCDmin = (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area)
                                                                        + (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["approach_landing_gear_out"].deltaCDgear = 0.03;
    theRaymerHighLiftAdaption["approach_landing_gear_out"].deltaCLatCDmin = -0.3;
    theRaymerHighLiftAdaption["landing"].deltaCDmin = (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theFrontWing), 0.33)) * (front_area/total_area)
                                                        + (0.15 * 1.0 / pow(geom2::measure::aspect_ratio(theBackWing), 0.33)) * (back_area/total_area);
    theRaymerHighLiftAdaption["landing"].deltaCDgear = 0.03;
    theRaymerHighLiftAdaption["landing"].deltaCLatCDmin = -0.4;

    return theRaymerHighLiftAdaption;
}

double tandemHighLiftAdaption::getAlphaAtCL(Polar aPolar, double aCL) {
    double AOAatACL = 0.0;
    size_t pointID(0);
    for (size_t i(0); i < aPolar.PolarPoints.size() - 1; ++i) {
        if (aPolar.PolarPoints.at(i).CL >= aCL) {
            AOAatACL = aPolar.PolarPoints.at(i).AoA;
            pointID = i;
            break;
        }
    }
    if (pointID != 0 && pointID != (aPolar.PolarPoints.size() - 1)) {
        // If required CL value is within the bounds and not at the edge --> Interpolate
        AOAatACL = aPolar.PolarPoints.at(pointID-1).AoA + (aPolar.PolarPoints.at(pointID).AoA
                    - aPolar.PolarPoints.at(pointID - 1).AoA)
                    / (aPolar.PolarPoints.at(pointID).CL - aPolar.PolarPoints.at(pointID - 1).CL)
                    * (aCL - aPolar.PolarPoints.at(pointID - 1).CL);
    }
    return AOAatACL;
}

double tandemHighLiftAdaption::getCLMax(Polar aPolar) {
    double CLMax = aPolar.PolarPoints.front().CL;
    for (size_t pointID(0); pointID < aPolar.PolarPoints.size(); ++pointID) {
        if (aPolar.PolarPoints.at(pointID).CL > CLMax) {
            CLMax = aPolar.PolarPoints.at(pointID).CL;
        }
    }
    return CLMax;
}
