/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CALCULATEPOLAR_SRC_METHODS_VISCDRAGFORTANDEM_H_
#define CALCULATEPOLAR_SRC_METHODS_VISCDRAGFORTANDEM_H_

#include <moduleBasics/runtimeIO.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/fuselage.h>

#include <memory>
#include <vector>
#include <string>

#include "../polar.h"

class viscDragForTandemInitialization {
 public:
    // geometry switches
    bool calculateFrontWing;                 /**< calculate viscous drag of the wing switch */
    bool calculateFuselage;             /**< calculate viscous drag of the fuselage switch */
    bool calculateVTP;                  /**< calculate viscous drag of the VTP switch */
    bool calculateBackWing;                  /**< calcualte viscous drag of the  BackWing switch */
    bool calculateNacelles;             /**< calculate viscous drag of the nacelles switch */
    bool calculateMisc;                 /**< calculate miscellaneous drag switch */
    // transition parameters
    bool manualTransitionLocation;          /**< switch wether transition location is given manually or flow is considered turbulent over all surfaces */
    double transitionLocationFrontWing;          /**< relative chord position on wing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationBackWing;           /**< relative chord position on BackWing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationVTP;           /**< relative chord position on VTP, where the flow goes from laminar to turbulent [-] */
    // calibration parameters
    bool doCalibration;
    double CDvisc_CDsum;
    double CDvisc_CLfact;
    double CDvisc_CLexp;
    double CDvisc_CDsum_lowMa;
    double CDvisc_CLfact_lowMa;
    double CDvisc_CLexp_lowMa;
    // other
    bool jetEngine;                         /**< switch if aircraft flies with jet engines or propellers */
    // geometry
    geom2::MultisectionSurface<geom2::AirfoilSection> theFrontWing;                      /**< geometry of the main wing */
    geom2::MultisectionSurface<geom2::AirfoilSection> theBackWing;                       /**< geometry of the horzontal tail plane */
    geom2::MultisectionSurface<geom2::AirfoilSection> theVTP;                       /**< geometry of the vertical tail plaine */
    geom2::MultisectionSurface<geom2::PolygonSection> theFuselage;                  /**< geometry of the fuselage */
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> theNacelles;     /**< vector containing all nacelle geometries */
    viscDragForTandemInitialization();
    virtual ~viscDragForTandemInitialization() {}
};
class viscDragForTandem {
 public:
    // general variables
    std::string config;
    std::filesystem::path aircraftPath;
    std::string airfoilPath;
    // geometry switches
    bool calculateFrontWing;                 /**< calculate viscous drag of the wing switch */
    bool calculateFuselage;             /**< calculate viscous drag of the fuselage switch */
    bool calculateVTP;                  /**< calculate viscous drag of the VTP switch */
    bool calculateBackWing;                  /**< calcualte viscous drag of the  BackWing switch */
    bool calculateNacelles;             /**< calculate viscous drag of the nacelles switch */
    bool calculateMisc;                 /**< calculate miscellaneous drag switch */
    // transition parameters
    bool manualTransitionLocation;
    double transitionLocationFrontWing;          /**< relative chord position on wing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationBackWing;    /**< relative chord position on BackWing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationVTP;           /**< relative chord position on VTP, where the flow goes from laminar to turbulent [-] */
    // calibration parameters
    bool doCalibration;
    double CDvisc_CDsum;
    double CDvisc_CLfact;
    double CDvisc_CLexp;
    double CDvisc_CDsum_lowMa;
    double CDvisc_CLfact_lowMa;
    double CDvisc_CLexp_lowMa;
    // sonstiges
    bool jetEngine;                         /**< switch if aircraft flies with jet engines or propellers */
    // geometry
    struct aircraftComponents {                                                         /**< struct which stores all possible aircraft geometry elements */
        geom2::MultisectionSurface<geom2::AirfoilSection> myFrontWing;                      /**< geometry of the main wing */
        geom2::MultisectionSurface<geom2::AirfoilSection> myBackWing;                       /**< geometry of the horzontal tail plane */
        geom2::MultisectionSurface<geom2::AirfoilSection> myVTP;                       /**< geometry of the vertical tail plaine */
        geom2::MultisectionSurface<geom2::PolygonSection> myFuselage;                  /**< geometry of the fuselage */
        std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> myNacelles;     /**< vector containing all nacelle geometries */
    };
    aircraftComponents theAircraftComponents;
    // geometry dependent wing parameters
    double Sref_FrontWing;           /**< Reference area of the wing [m^2] */
    double Swet_FrontWing;           /**< Wetted area of the wing [m^2] */
    double MAC_FrontWing;            /**< Mean aerodynamic chord of the wing [m] */
    double avgTtoC_FrontWing;        /**< Average thickness to chord ratio of the wing [-] */
    double eta_ref_FrontWing;        /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_FrontWing;     /**< Local sweep at reference position [rad] */
    double mXtoC_FrontWing;          /**< Chordwise location of airfoil maximum thickness point [-] */
    double QC_FrontWing;             /**< Form factor wing [-] */
    // geometry dependent BackWing parameters
    double Sref_BackWing;           /**< Reference area of the back wing [m^2] */
    double Swet_BackWing;            /**< Wetted area of the BackWing [m^2] */       //TODO kayra: buradaki parametreleri de degistirmem lazim mi?
    double MAC_BackWing;             /**< Mean aerodynamic chord of the BackWing [m] */
    double avgTtoC_BackWing;         /**< Average thickness to chord ratio of the BackWing [-] */
    double mXtoC_BackWing;           /**< Chordwise location of airfoil maximum thickness point [-] */
    double eta_ref_BackWing;         /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_BackWing;      /**< Local sweep at reference position [rad] */
    double QC_BackWing;              /**< Form factor BackWing [-] */
    // geometry dependent VTP parameters
    double Swet_VTP;            /**< Wetted area of the VTP [m^2] */
    double MAC_VTP;             /**< Mean aerodynamic chord of the VTP [m] */
    double avgTtoC_VTP;         /**< Average thickness to chord ratio of the VTP [-] */
    double mXtoC_VTP;           /**< Chordwise location of airfoil maximum thickness point [-] */
    double eta_ref_VTP;         /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_VTP;      /**< Local sweep at reference position [rad] */
    double QC_VTP;              /**< Form factor VTP [-] */
    // geometry dependent fuselage parameters
    double Swet_Fuselage;       /**< Wetted area of the fuselage [m^2] */
    double l_Fuselage;          /**< lenght of the fuselage [m] */
    double maxWidthFuselage;    /**< maximum width of the fuselage [m] */
    double maxHeightFuselage;   /**< maximum height of the fuselage [m] */
    double QC_Fuselage;         /**< Form factor fuselage [-] */
    double fuselageUpsweepAngle;  /**< Up sweep angle of the fuselage tail [rad] */
    // geometry dependent nacelle parameters
    double Swet_Nacelles;       /**< Wetted area of one nacelle [m^2] */
    double l_Nacelles;          /**< Length of one nacelle [m] */
    double maxWidthNacelles;    /**< Maximum width of one nacelle [m] */
    double maxHeightNacelles;   /**< Maximum height of one nacelle [m] */
    double QC_Nacelles;         /**< Form factor nacelles [-] */
    double factorProtuberances; /**< protuberance factor of the aircraft [-] */
    // calculate parameters that are only dependent from the geometry
    void preprocessGeometry();
    // calculations
    /** \brief Overall function to calculate viscous drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationFrontWing relative transition location where the flow turns turbulent on the wing
     *  \param transitionLocationBackWing relative transition location where the flow turns turbulent on the BackWing
     *  \param transitionLocationVTP relative transition location where the flow turns turbulent on the VTP
     */   
    std::vector <double> calculateViscDrag(const Polar& aPolar, const double transitionLocationFrontWing,
                                    const double transitionLocationBackWing, const double transitionLocationVTP);
    /** \brief Overall function to calculate viscous drag corrections
     *  \param componentDrags vector of the component drags which need to be corrected
     *  \param aMachNumber mach number for the calibration
     *  \param CL CL value for the calibration
     */                                       
    std::vector <double> calculateCorrectedViscDrag(const std::vector <double>& componentDrags, const double aMachNumber, const double CL);
    /** \brief Function to calculate the wing component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationFrontWing relative transition location where the flow turns turbulent on the wing
     */          
    double calculateViscDragFrontWing(const Polar& aPolar, const double transitionLocationFrontWing);
    /** \brief Function to calculate the BackWing component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationBackWing relative transition location where the flow turns turbulent on the BackWing
     */  
    double calculateViscDragBackWing(const Polar& aPolar, const double transitionLocationBackWing);
    /** \brief Function to calculate the VTP component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationVTP relative transition location where the flow turns turbulent on the VTP
     */  
    double calculateViscDragVTP(const Polar& aPolar, const double transitionLocationVTP);
    /** \brief Function to calculate the fuselage component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */  
    double calculateViscDragFuselage(const Polar& aPolar);
    /** \brief Function to calculate the nacelle component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */  
    double calculateViscDragNacelles(const Polar& aPolar);
    /** \brief Function to calculate the miscellaneous component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */     
    double calculateViscDragMisc(const Polar& aPolar);
    /** \brief Function to calculate the friction coefficient
     *  \param flightMach mach number
     *  \param ReynoldsNumber reynolds number
     */ 
    double getFrictionCoefficient(double flightMach, double ReynoldsNumber);
    explicit viscDragForTandem(const std::shared_ptr<RuntimeIO>& rtIO, const viscDragForTandemInitialization& viscDragInit);
    virtual ~viscDragForTandem() {}
};

#endif  // CALCULATEPOLAR_SRC_METHODS_VISCDRAGFORTANDEM_H_
