/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tandemDefaultStrategy.h"

#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <unitConversion/unitConversion.h>
#include <CGAL/Surface_mesh/IO/PLY.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <memory>
#include <vector>
#include <filesystem>
#include <utility>  // make_pair

#include "../polar.h"
#include "../methods/generalMethods/auxFunctions.h"
#include "../methods/liftingLineForTandem.h" 
#include "../methods/viscDragForTandem.h"
#include "../methods/waveDragMason.h"
#include "../methods/tandemHighLiftAdaption.h"
#include "../methods/trimInterpolation.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::pair;
using std::vector;

/*  TODO:   - check if acft xml has entries before building the geometry comopnents */
tandemDefaultStrategy::tandemDefaultStrategy(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<tandemCalculatePolarConfig>& config)
    :
    CoGnames{"design", "front", "aft"},
    HLconfigs{"takeoff", "takeoff_landing_gear_retracted", "climb", "approach", "approach_landing_gear_out", "landing"},
    C_LmaxLanding(0.0),
    C_LmaxTO(0.0),
    C_LoptimumCruise(0.0),
    rtIO_(rtIO),
    data_(std::make_shared<tandemDefaultData>(rtIO_)),
    config_(config) {
}

void tandemDefaultStrategy::run() {     
    // loop over all cg positions and back wing angles
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t backWingID(0); backWingID < config_->backWingAngles.size(); ++backWingID) {
            this->clearDataForNextCalculationRun();
            myRuntimeInfo-> out << endl;
            myRuntimeInfo->out << "Run aero calculations for " << config_->CGPositions.at(CGID) <<
                            " CoG and back wing angle of " << config_->backWingAngles.at(backWingID) << " deg..." << endl;
            myRuntimeInfo-> out << endl;
            this->createAircraftGeometry(config_->backWingAngles.at(backWingID)); 
            this->runAeroCalculations(data_->CoGPositions.at(config_->CGPositions.at(CGID)), config_->backWingAngles.at(backWingID));
            this->copyPolarsToCGMap(CGID, backWingID);
        }
    }
    // interpolate trimed polars, or just copy polars if trim is disabled
    if (config_->doTrim.value()) {
        this->doTrimInterpolation();
        if (config_->trimHighLiftPolars.value()) {
            this->doTrimInterpolationHighLift();
        } else {
            this->copyHighLiftPolars();
        }
    } else {
        this->copyCleanPolars();
        this->copyHighLiftPolars();
    }
    this->calibratePolars();
}

void tandemDefaultStrategy::createAircraftGeometry(const double backWingAngle) {
    auto get_position = [](const node &component) -> geom2::Point_3 {
        return geom2::Point_3{double{component.at("position/x/value")},
                              double{component.at("position/y/value")},
                              double{component.at("position/z/value")}};
    };
    std::filesystem::path file{rtIO_->acxml.getName()};
    std::shared_ptr<node> AcXML = aixml::openDocument(file);
    // create wings
    this->aircraft_xml = aixml::openDocument(rtIO_->acxmlAccess);
    const auto tandem_wing_set_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/wing"));
    geom2::WingFactory wingFactory{AcXML, rtIO_->getAirfoilDataDir()};
    // front wing
    geom2::MultisectionSurface<geom2::AirfoilSection> frontWing = wingFactory.create("wing/specific/geometry/aerodynamic_surface@0");
    auto offset = tandem_wing_set_position - CGAL::ORIGIN;
    frontWing.origin += offset;
    liftingSurfaces.push_back(frontWing);
    // back wing
    geom2::MultisectionSurface<geom2::AirfoilSection> backWing = wingFactory.create("wing/specific/geometry/aerodynamic_surface@1");
    backWing.origin += offset; 
    for (auto& section : backWing.sections) {       //TODO kayra: 0 section da dönsün mü?
        double designBackWingAngle = section.get_twist_angle();
        section.set_twist_angle(designBackWingAngle + convertUnit(DEGREE, RADIAN, backWingAngle));
    }
    liftingSurfaces.push_back(backWing);
    // create fuselage
    geom2::FuselageFactory fuselageFactory{AcXML, rtIO_->getGeometryDir()};
    fuselage = fuselageFactory.create("fuselage/specific/geometry/fuselage@0");
    // create VTP
    const auto vtp_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/empennage"));
    geom2::WingFactory VTPFactory{AcXML, rtIO_->getAirfoilDataDir()};
    VTP = VTPFactory.create("empennage/specific/geometry/aerodynamic_surface@0");
    offset = vtp_position - CGAL::ORIGIN;
    VTP.origin += offset;
    geom2::Mesh frontWing_mesh = geom2::transform::to_mesh(frontWing);
    geom2::Mesh backWing_mesh  = geom2::transform::to_mesh(backWing);
    backWing_mesh += frontWing_mesh;
    geom2::Mesh vtp_mesh = geom2::transform::to_mesh(VTP);
    vtp_mesh += backWing_mesh;
    geom2::Mesh fuselage_mesh = geom2::transform::to_mesh(fuselage);
    fuselage_mesh += vtp_mesh;
    // create nacelles
    const auto nacelle_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/propulsion"));
    offset = nacelle_position - CGAL::ORIGIN;
    size_t numberOfPropulsors = rtIO_->acxml.getVector(("aircraft_exchange_file/component_design/propulsion/specific/propulsion"), 1).size();
    geom2::HullFactory nacelleFactory{AcXML, rtIO_->getGeometryDir()};
    for (size_t i(0); i < numberOfPropulsors; ++i) {
        geom2::MultisectionSurface<geom2::PolygonSection> aNacelle = nacelleFactory.create("propulsion/specific/propulsion@" + num2Str(i) + "/nacelle");
        aNacelle.origin += offset;
        geom2::Mesh aNacelle_mesh = geom2::transform::to_mesh(aNacelle);
        fuselage_mesh += aNacelle_mesh;
        nacelles.push_back(aNacelle);
    }
    CGAL::IO::write_STL("tandemWing_mesh.stl", fuselage_mesh); 
}

void tandemDefaultStrategy::runAeroCalculations(const Point CoGPosition, const double currentIBackWingAngle) {
    // lifting line for tandem
    LiLiforTandemInitialization myLiLiInit;
    this->initializeLiLiforTandem(&myLiLiInit);
    liftingLineForTandem* myLiftingLineForTandem = new liftingLineForTandem(rtIO_, config_->liftingLineFolderPath, myLiLiInit);
    myLiftingLineForTandem->run(CoGPosition);
    this->getPolarDataFromLiftingLine(*myLiftingLineForTandem);
    // viscous drag by raymer
    viscDragForTandemInitialization myViscDragInit;
    this->initializeViscDragForTandem(&myViscDragInit);
    viscDragForTandem* myViscDragForTandem = new viscDragForTandem(rtIO_, myViscDragInit);
    this->calculateViscDragForTandem(myViscDragForTandem);
    // wave drag by mason
    waveDragInitialization myWaveDragInit;
    this->initializeWaveDragMason(&myWaveDragInit);
    waveDragMason* myWaveDragMason = new waveDragMason(rtIO_, myWaveDragInit);
    this->calculateWaveDragMason(myWaveDragMason);
    // summation of the drag components
    this->sumUpDragComponents();
    // high lift calculation
    tandemHighLiftAdaption* myTandemHLAdaption = new tandemHighLiftAdaption(rtIO_, liftingSurfaces.at(0), liftingSurfaces.at(1));
    this->copyWingDevices(myTandemHLAdaption);
    this->copyCleanPolarForHLCalculation(myTandemHLAdaption, myLiftingLineForTandem);
    myTandemHLAdaption->runHighLiftAdaption();
    this->calculateHighLiftPolars(myTandemHLAdaption, CoGPosition.xCoordinate, currentIBackWingAngle);
}

void tandemDefaultStrategy::initializeLiLiforTandem(LiLiforTandemInitialization *myLiLiInit) {
    myLiLiInit->CLModeAirfoilSelection = config_->CLModeAirfoilSelection.value();
    myLiLiInit->stepWidthCL = config_->stepWidthCL.value();
    myLiLiInit->delta_CM0 = config_->delta_CM0.value();
    myLiLiInit->delta_dCMdCL = config_->delta_dCMdCL.value();
    myLiLiInit->indDragCtCorrForCalibration = config_->indDragCtCorrForCalibration.value();
    myLiLiInit->factorDragCleanPolar = config_->factorIndDragCleanPolar.value();
    myLiLiInit->bestCruiseCL = data_->bestCruiseCL;
    myLiLiInit->initialMachCruise = data_->initialCruiseMachNumber;
    //myLiLiInit->engineType = data_->engineType;
    for (size_t flightConditionID(0); flightConditionID < config_->flightConditions.size(); ++flightConditionID) {
        flightCondition* aLiLiFlightCondition = new flightCondition;
        aLiLiFlightCondition->Mach = config_->flightConditions.at(flightConditionID).Mach;
        aLiLiFlightCondition->Altitude = config_->flightConditions.at(flightConditionID).Altitude;
        aLiLiFlightCondition->extrapolationMargin = config_->flightConditions.at(flightConditionID).extrapolationMargin;
        aLiLiFlightCondition->allowGridChange = config_->flightConditions.at(flightConditionID).allowGridChange;
        myLiLiInit->flightConditions.push_back(*aLiLiFlightCondition);
        if (accuracyCheck(aLiLiFlightCondition->Mach, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            myLiLiInit->cruiseMachID = flightConditionID;
        }
        delete aLiLiFlightCondition;
    }
    myLiLiInit->liftingSurfaces = liftingSurfaces;
}

void tandemDefaultStrategy::initializeViscDragForTandem(viscDragForTandemInitialization *myViscDragInit) {
    // copy geometry
    myViscDragInit->theFrontWing = liftingSurfaces.at(0);
    myViscDragInit->calculateFrontWing = true;
    myViscDragInit->theBackWing = liftingSurfaces.at(1);
    myViscDragInit->calculateBackWing = true;
    myViscDragInit->theVTP = VTP;
    myViscDragInit->calculateVTP = true;
    myViscDragInit->theFuselage = fuselage;
    myViscDragInit->calculateFuselage = true;
    myViscDragInit->theNacelles = nacelles;
    myViscDragInit->calculateNacelles = true;
    myViscDragInit->calculateMisc = true;
    // initialize transition parameters
    myViscDragInit->manualTransitionLocation = config_->manualTransition.value();  
    myViscDragInit->transitionLocationFrontWing = config_->manualTransitionFrontWing.value();
    myViscDragInit->transitionLocationBackWing = config_->manualTransitionBackWing.value();
    myViscDragInit->transitionLocationVTP = config_->manualTransitionFin.value();
    // initialize calibration parameters
    myViscDragInit->doCalibration = config_->doViscDragCalibration.value();
    myViscDragInit->CDvisc_CDsum = config_->CDvisc_CDsum.value();
    myViscDragInit->CDvisc_CLexp = config_->CDvisc_CLexp.value();
    myViscDragInit->CDvisc_CLfact = config_->CDvisc_CLfact.value();
    myViscDragInit->CDvisc_CDsum_lowMa = config_->CDvisc_CDsum_lowMa.value();
    myViscDragInit->CDvisc_CLexp_lowMa = config_->CDvisc_CLexp_lowMa.value();
    myViscDragInit->CDvisc_CLfact_lowMa = config_->CDvisc_CLfact_lowMa.value();
    // calculate parameters that remain static over all polars
}

void tandemDefaultStrategy::initializeWaveDragMason(waveDragInitialization *myWaveDragInit) {
    // fill configuration parameters and values in the wave drag method
    myWaveDragInit->CLExp = config_->CDwave_CLExp.value();
    myWaveDragInit->CLFact = config_->CDwave_CLFact.value();
    myWaveDragInit->factorKa = config_->Ka.value();
    myWaveDragInit->initialMachCruise = data_->initialCruiseMachNumber;
    myWaveDragInit->numberOfSegments = config_->numberWingStrips.value();
    myWaveDragInit->waveDragCtCorrForCalibration = config_->deltaWaveDragCountWing.value();
    myWaveDragInit->maxAllowedSweep = convertUnit(RADIAN, DEGREE, config_->maximumSegmentSweep.value());
    myWaveDragInit->wFuselage = {geom2::measure::width(fuselage, liftingSurfaces.at(0).origin.x()) * 0.5, geom2::measure::width(fuselage, liftingSurfaces.at(1).origin.x()) * 0.5};  
    myWaveDragInit->liftingSurfaces = liftingSurfaces;
}

void tandemDefaultStrategy::getPolarDataFromLiftingLine(const liftingLineForTandem& liftingLineForTandem) {
    for (size_t machID(0); machID < config_->flightConditions.size(); ++machID) {
        tandemDefaultPolar aCleanPolar;
        aCleanPolar.extrapolationMargin = liftingLineForTandem.theCleanPolars.at(machID).extrapolationMargin;
        aCleanPolar.allowGridChange = liftingLineForTandem.theCleanPolars.at(machID).allowGridChange;
        aCleanPolar.altitude = liftingLineForTandem.theCleanPolars.at(machID).altitude;
        aCleanPolar.MachNumber = liftingLineForTandem.theCleanPolars.at(machID).MachNumber;
        aCleanPolar.ReynoldsNumber = liftingLineForTandem.theCleanPolars.at(machID).ReynoldsNumber;  //TODO kayra: sadece arka kanada göre yapiyor  ortalama MAc mi alsak ve ref area sadece ön
        for (size_t pointID(0); pointID < liftingLineForTandem.theCleanPolars.at(machID).Points.size(); ++pointID) {
            aCleanPolar.Points.push_back(defaultPolarPoint());
            aCleanPolar.Points.back().CL = liftingLineForTandem.theCleanPolars.at(machID).Points.at(pointID).CL;
            aCleanPolar.Points.back().AoA = liftingLineForTandem.theCleanPolars.at(machID).Points.at(pointID).AoA;
            aCleanPolar.Points.back().CD_ind = liftingLineForTandem.theCleanPolars.at(machID).Points.at(pointID).CDind;
            aCleanPolar.Points.back().CM = liftingLineForTandem.theCleanPolars.at(machID).Points.at(pointID).CM;
            aCleanPolar.Points.back().iStabPolar = liftingLineForTandem.theCleanPolars.at(machID).Points.at(pointID).iStabPolar;   //TODO kayra: holds back wing angle for tandem
            aCleanPolar.PointsFrontWing.push_back(defaultPolarPointLiftSurface());
            aCleanPolar.PointsFrontWing.back().yStations = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).yStations;
            aCleanPolar.PointsFrontWing.back().Cl_distr = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Cl_distr;
            aCleanPolar.PointsFrontWing.back().Lift_distr = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Lift_distr;
            aCleanPolar.PointsFrontWing.back().CL = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CL;
            aCleanPolar.PointsFrontWing.back().CD_ind = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CDind;
            aCleanPolar.PointsBackWing.push_back(defaultPolarPointLiftSurface());
            aCleanPolar.PointsBackWing.back().yStations = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).yStations;
            aCleanPolar.PointsBackWing.back().Cl_distr = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).Cl_distr;
            aCleanPolar.PointsBackWing.back().Lift_distr = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).Lift_distr;
            aCleanPolar.PointsBackWing.back().CL = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).CL;
            aCleanPolar.PointsBackWing.back().CD_ind = liftingLineForTandem.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).CDind;
            //TODO kayra:Even though liftingLine has no fuselage/nacelle data, i create an entry so that we can store our Raymer results later
            aCleanPolar.PointsNacelles.push_back(defaultPolarPoint());   
            aCleanPolar.PointsFuselage.push_back(defaultPolarPoint());
            aCleanPolar.PointsFin.push_back(defaultPolarPoint());
            aCleanPolar.PointsMisc.push_back(defaultPolarPoint());
        }
        currentPolarSet.cleanPolars.insert(std::make_pair(machID, aCleanPolar));
    }
    LILI_Version = liftingLineForTandem.LILI_version;
}

void tandemDefaultStrategy::calculateViscDragForTandem(viscDragForTandem *myViscDragForTandem) {
    myViscDragForTandem->preprocessGeometry();
    // run viscous drag calculation by Raymer over the current polar set
    for (std::pair<size_t, tandemDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        size_t polarID = polarIterator.first;
        vector<double> theComponentDrags = myViscDragForTandem->calculateViscDrag(currentPolarSet.cleanPolars.at(polarID), 0.0, 0.0, 0.0); 
        // loop over all points to calculate CL dependent visc drag correction                                                               
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            vector <double> theCorrectedComponentDrags = myViscDragForTandem->calculateCorrectedViscDrag(theComponentDrags, currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                                        currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CL); 
            // assign total and componentwise viscous drag to the polar map
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc = std::accumulate(theCorrectedComponentDrags.begin(), theCorrectedComponentDrags.end(), 0.0);
            currentPolarSet.cleanPolars.at(polarID).PointsFrontWing.at(pointID).CD_visc = theCorrectedComponentDrags.at(0);
            currentPolarSet.cleanPolars.at(polarID).PointsBackWing.at(pointID).CD_visc = theCorrectedComponentDrags.at(1);
            currentPolarSet.cleanPolars.at(polarID).PointsFin.at(pointID).CD_visc = theCorrectedComponentDrags.at(2);
            currentPolarSet.cleanPolars.at(polarID).PointsFuselage.at(pointID).CD_visc = theCorrectedComponentDrags.at(3);
            currentPolarSet.cleanPolars.at(polarID).PointsNacelles.at(pointID).CD_visc = theCorrectedComponentDrags.at(4);
            currentPolarSet.cleanPolars.at(polarID).PointsMisc.at(pointID).CD_visc = theCorrectedComponentDrags.at(5);
        }
    }
    myRuntimeInfo->out<< "... Viscous drag has been calculated sucessfully!" << endl;
}

void tandemDefaultStrategy::calculateWaveDragMason(waveDragMason *myWaveDragMason) {
    double frontWingArea = geom2::measure::reference_area(liftingSurfaces.at(0));
    double backWingArea = geom2::measure::reference_area(liftingSurfaces.at(1));
    double totalWingArea = frontWingArea + backWingArea;
    // run wave drag calculation by Mason over the current polar set
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        double sweepPosition(0.0);
        if (currentPolarSet.cleanPolars.at(polarID).MachNumber < 0.7) {
            sweepPosition = 0.25;
        } else {
            sweepPosition = 0.5;
        }
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).PointsFrontWing.at(pointID).CD_wave
                                            = myWaveDragMason->calculateWaveDrag(currentPolarSet.cleanPolars.at(polarID).Points.at(pointID),
                                                                                currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                0, sweepPosition, currentPolarSet.cleanPolars.at(polarID).PointsFrontWing.at(pointID).yStations,
                                                                                currentPolarSet.cleanPolars.at(polarID).PointsFrontWing.at(pointID).Cl_distr, false) * (frontWingArea / totalWingArea); //TODO kayra: false corrected
            currentPolarSet.cleanPolars.at(polarID).PointsBackWing.at(pointID).CD_wave                                                                                                                  
                                            = myWaveDragMason->calculateWaveDrag(currentPolarSet.cleanPolars.at(polarID).Points.at(pointID),
                                                                                currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                1, sweepPosition, currentPolarSet.cleanPolars.at(polarID).PointsBackWing.at(pointID).yStations,
                                                                                currentPolarSet.cleanPolars.at(polarID).PointsBackWing.at(pointID).Cl_distr, false) * (backWingArea / totalWingArea);
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave = currentPolarSet.cleanPolars.at(polarID).PointsFrontWing.at(pointID).CD_wave  
                                                                               + currentPolarSet.cleanPolars.at(polarID).PointsBackWing.at(pointID).CD_wave;   
        }
    }
    myRuntimeInfo->out << "... Wave drag has been calculated sucessfully!" << endl;
}

void tandemDefaultStrategy::sumUpDragComponents() {         
    myRuntimeInfo->out << "Sum up drag components" << endl;
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD =  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_ind
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave;
        }
    }
}

void tandemDefaultStrategy::copyWingDevices(tandemHighLiftAdaption *myTandemHLAdaption) {
    myRuntimeInfo->out << "Copy wing devices" << endl;
    for (size_t deviceID(0); deviceID < data_->front_numberDevices; ++deviceID) {
        tandemHighLiftAdaption::WingDevice* aWingDevice = new tandemHighLiftAdaption::WingDevice;
        aWingDevice->name = data_->front_wingDevices.at(deviceID).type;
        aWingDevice->innerRelFwdChordPosition = data_->front_wingDevices.at(deviceID).innerRelFwdChordPosition;
        aWingDevice->innerRelAftChordPosition = data_->front_wingDevices.at(deviceID).innerRelAftChordPosition;
        aWingDevice->outerRelFwdChordPosition = data_->front_wingDevices.at(deviceID).outerRelFwdChordPosition;
        aWingDevice->outerRelAftChordPosition = data_->front_wingDevices.at(deviceID).outerRelAftChordPosition;
        aWingDevice->innerSpan_rel = data_->front_wingDevices.at(deviceID).innerRelSpanPosition;
        aWingDevice->outerSpan_rel = data_->front_wingDevices.at(deviceID).outerRelSpanPosition;
        myTandemHLAdaption->front_wingDevices.push_back(*aWingDevice);
    }
    for (size_t deviceID(0); deviceID < data_->back_numberDevices; ++deviceID) {
        tandemHighLiftAdaption::WingDevice* aWingDevice = new tandemHighLiftAdaption::WingDevice;
        aWingDevice->name = data_->back_wingDevices.at(deviceID).type;
        aWingDevice->innerRelFwdChordPosition = data_->back_wingDevices.at(deviceID).innerRelFwdChordPosition;
        aWingDevice->innerRelAftChordPosition = data_->back_wingDevices.at(deviceID).innerRelAftChordPosition;
        aWingDevice->outerRelFwdChordPosition = data_->back_wingDevices.at(deviceID).outerRelFwdChordPosition;
        aWingDevice->outerRelAftChordPosition = data_->back_wingDevices.at(deviceID).outerRelAftChordPosition;
        aWingDevice->innerSpan_rel = data_->back_wingDevices.at(deviceID).innerRelSpanPosition;
        aWingDevice->outerSpan_rel = data_->back_wingDevices.at(deviceID).outerRelSpanPosition;
        myTandemHLAdaption->back_wingDevices.push_back(*aWingDevice);
    }
}

void tandemDefaultStrategy::copyCleanPolarForHLCalculation(tandemHighLiftAdaption *myTandemHLAdaption, liftingLineForTandem *myLiftingLineForTandem) {
    // find polar with minimum mach number
    double minMachNumber(1.0);
    size_t minMachID;
    for (std::pair<size_t, tandemDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        if (polarIterator.second.MachNumber < minMachNumber) {
            minMachID = polarIterator.first;
            minMachNumber = polarIterator.second.MachNumber;
        }
    }
    // copy clean polar to HL method
    myTandemHLAdaption->theCleanPolar.MachNumber = currentPolarSet.cleanPolars.at(minMachID).MachNumber;
    myTandemHLAdaption->theCleanPolar.altitude = currentPolarSet.cleanPolars.at(minMachID).altitude;
    myTandemHLAdaption->theCleanPolar.extrapolationMargin = currentPolarSet.cleanPolars.at(minMachID).extrapolationMargin;
    myTandemHLAdaption->theCleanPolar.allowGridChange = currentPolarSet.cleanPolars.at(minMachID).allowGridChange;
    myTandemHLAdaption->theCleanPolar.ReynoldsNumber = currentPolarSet.cleanPolars.at(minMachID).ReynoldsNumber;
    myTandemHLAdaption->theCleanPolar.Config = currentPolarSet.cleanPolars.at(minMachID).Config;
    for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(minMachID).Points.size(); ++pointID) {
        myTandemHLAdaption->theCleanPolar.PolarPoints.push_back(PolarPoint());
        myTandemHLAdaption->theCleanPolar.PolarPoints.back().AoA = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).AoA;
        myTandemHLAdaption->theCleanPolar.PolarPoints.back().CL = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CL;
        myTandemHLAdaption->theCleanPolar.PolarPoints.back().CD = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CD;
        myTandemHLAdaption->theCleanPolar.PolarPoints.back().CM = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CM;
        myTandemHLAdaption->theCleanPolar.PolarPoints.back().iStabPolar = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).iStabPolar;
    }
    // copy other neccesary values from lifting line calculation
    myTandemHLAdaption->CLmaxTO_req = 1.0;
    myTandemHLAdaption->CLmaxLDG_req = 1.0;
    myTandemHLAdaption->stepWidthCL = config_->stepWidthCL.value();
    myTandemHLAdaption->reductionDragCountsHighLift = config_->deltaTotalDragHL.value();
    myTandemHLAdaption->factorDragHighLift = config_->factorDragHL.value();
    myTandemHLAdaption->CDmin_inc = myLiftingLineForTandem->theAnalyticalPolarData.CDmin_inc.at(0);  
    myTandemHLAdaption->kFactor_inc = myLiftingLineForTandem->theAnalyticalPolarData.kFactor_inc.at(0);
    myTandemHLAdaption->CLatCDmin_inc = myLiftingLineForTandem->theAnalyticalPolarData.CLatCDmin_inc.at(0);
    myTandemHLAdaption->dCMtodCL_inc = myLiftingLineForTandem->theAnalyticalPolarData.dCMtodCL_inc.at(0);
    myTandemHLAdaption->CMatCL0_inc = myLiftingLineForTandem->theAnalyticalPolarData.CMatCL0_inc.at(0);
    myTandemHLAdaption->dCMtodAoA = myLiftingLineForTandem->theAnalyticalPolarData.dCMtodAoA.at(0);
    myTandemHLAdaption->dCLtodAoA = myLiftingLineForTandem->theAnalyticalPolarData.dCLtodAoA.at(0);
    myTandemHLAdaption->CLatAoA0 = myLiftingLineForTandem->theAnalyticalPolarData.CLatAoA0.at(0);
    myTandemHLAdaption->CMatAoA0 = myLiftingLineForTandem->theAnalyticalPolarData.CMatAoA0.at(0);
}

void tandemDefaultStrategy::calculateHighLiftPolars(tandemHighLiftAdaption *myTandemHLAdaption, const double CoG_X_Position, const double currentIBackWingAngle) {
    for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
        myRuntimeInfo->out << "Calculate high lift polar for " << HLconfigs.at(configID) << endl;
        Polar aPolar = myTandemHLAdaption->calculateHighLiftPolar(HLconfigs.at(configID), CoG_X_Position, currentIBackWingAngle);
        currentPolarSet.highLiftPolars[HLconfigs.at(configID)] = tandemDefaultPolar();
        // copy polar data
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Config = HLconfigs.at(configID);
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).extrapolationMargin = aPolar.extrapolationMargin;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).allowGridChange = aPolar.allowGridChange;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).altitude = aPolar.altitude;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).MachNumber = aPolar.MachNumber;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).ReynoldsNumber = aPolar.ReynoldsNumber;
        for (size_t pointID(0); pointID < aPolar.PolarPoints.size(); ++pointID) {
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.push_back(defaultPolarPoint());
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().AoA = aPolar.PolarPoints.at(pointID).AoA;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CL = aPolar.PolarPoints.at(pointID).CL;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CD = aPolar.PolarPoints.at(pointID).CD; 
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CM = aPolar.PolarPoints.at(pointID).CM;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().iStabPolar = aPolar.PolarPoints.at(pointID).iStabPolar;
        }
    }
}

void tandemDefaultStrategy::copyPolarsToCGMap(const size_t CGID, const size_t backWingID) {
    polarDataCGMap[CGID][backWingID] = currentPolarSet;
}

void tandemDefaultStrategy::clearDataForNextCalculationRun() {
    currentPolarSet.cleanPolars.clear();
    currentPolarSet.highLiftPolars.clear();
    liftingSurfaces.clear();
}

void tandemDefaultStrategy::doTrimInterpolation() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        vector<vector<trimInterpolation::interpolationPoint>> polarSet;
        // copy polar data into the generalized interpolation container
        for (size_t polarID(0); polarID < polarDataCGMap.at(CGID).at(0).cleanPolars.size(); ++polarID) {
            myRuntimeInfo->out << "Interpolate trim for clean polars at " << config_->CGPositions.at(CGID) << " CoG at Ma " << config_->machNumbers.at(polarID) << "..." << endl;
            for (size_t backWingID(0); backWingID < config_->backWingAngles.size(); ++backWingID) {
                vector<trimInterpolation::interpolationPoint> aBackWingPolar;
                for (size_t pointID(0); pointID < polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.size(); ++pointID) {
                    trimInterpolation::interpolationPoint anInterpolationPoint;
                    anInterpolationPoint.CL = polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CL;
                    anInterpolationPoint.CM = polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CM;
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CD);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).AoA);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).iStabPolar);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CD_ind);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CD_visc);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).Points.at(pointID).CD_wave);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsFrontWing.at(pointID).CL);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).CL);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsFrontWing.at(pointID).CD_wave);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).CD_ind);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).CD_wave);
                    // interpolate additional values for cruise mach polar
                    if (accuracyCheck(data_->initialCruiseMachNumber, polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).MachNumber, ACCURACY_HIGH)) {
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsFrontWing.at(pointID).Cl_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsFrontWing.at(pointID).Lift_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsFrontWing.at(pointID).yStations);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).Cl_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).Lift_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(polarID).PointsBackWing.at(pointID).yStations);
                    }
                    aBackWingPolar.push_back(anInterpolationPoint);
                }
                polarSet.push_back(aBackWingPolar);
            }
            // interpolate the current polar set
            trimInterpolation* myTrimInterpolation = new trimInterpolation();
            vector<trimInterpolation::interpolationPoint> interpolatedTrimPolar = myTrimInterpolation->interpolateTrim(polarSet, true);
            polarSet.clear();
            // copy data to polarDataTrimMap
            polarDataTrimMap[CGID].cleanPolars[polarID].MachNumber = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).MachNumber;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).ReynoldsNumber = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).ReynoldsNumber;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).altitude = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).altitude;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).extrapolationMargin = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).extrapolationMargin;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).allowGridChange = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).allowGridChange;
            for (size_t pointID(0); pointID < interpolatedTrimPolar.size(); ++pointID) {
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.push_back(defaultPolarPoint());
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsFrontWing.push_back(defaultPolarPointLiftSurface());
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsBackWing.push_back(defaultPolarPointLiftSurface());
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CL = interpolatedTrimPolar.at(pointID).CL;
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CM = interpolatedTrimPolar.at(pointID).CM;
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CD = interpolatedTrimPolar.at(pointID).values.at(0);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().AoA = interpolatedTrimPolar.at(pointID).values.at(1);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().iStabPolar = interpolatedTrimPolar.at(pointID).values.at(2);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CD_ind = interpolatedTrimPolar.at(pointID).values.at(3);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CD_visc = interpolatedTrimPolar.at(pointID).values.at(4);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CD_wave = interpolatedTrimPolar.at(pointID).values.at(5);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsFrontWing.back().CL = interpolatedTrimPolar.at(pointID).values.at(6);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsBackWing.back().CL = interpolatedTrimPolar.at(pointID).values.at(7);
                if (accuracyCheck(data_->initialCruiseMachNumber, polarDataTrimMap.at(CGID).cleanPolars.at(polarID).MachNumber, ACCURACY_HIGH)) {
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsFrontWing.back().Cl_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(0);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsFrontWing.back().Lift_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(1);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsFrontWing.back().yStations = interpolatedTrimPolar.at(pointID).vectorValues.at(2);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsBackWing.back().Cl_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(3);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsBackWing.back().Lift_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(4);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsBackWing.back().yStations = interpolatedTrimPolar.at(pointID).vectorValues.at(5);
                }
            }
        }
    }
    myRuntimeInfo->out << "All polars are trimmed!" << endl;
}

void tandemDefaultStrategy::doTrimInterpolationHighLift() {
    // TBA
}

void tandemDefaultStrategy::copyCleanPolars() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t polarID(0); polarID < polarDataCGMap.at(CGID).at(0).cleanPolars.size(); ++polarID) {
            polarDataTrimMap[CGID].cleanPolars[polarID] = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID);
        }
    }
}

void tandemDefaultStrategy::copyHighLiftPolars() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
            polarDataTrimMap[CGID].highLiftPolars[HLconfigs.at(configID)] = polarDataCGMap.at(CGID).at(config_->neutralBackWingAngleID).highLiftPolars.at(HLconfigs.at(configID));
        }
    }
}

void tandemDefaultStrategy::calibratePolars() {
    if (!(accuracyCheck(config_->factorDragCleanPolar.value(), 1.0, ACCURACY_HIGH) && accuracyCheck(config_->deltaTotalDragCleanPolar.value(), 0.0, ACCURACY_HIGH))) {
        calibrateCleanPolars();
    }
    if (!(accuracyCheck(config_->factorDragHighLiftPolar.value(), 1.0, ACCURACY_HIGH) && accuracyCheck(config_->deltaTotalDragHighLift.value(), 0.0, ACCURACY_HIGH))) {
        calibrateHighLiftPolars();
    }
}

void tandemDefaultStrategy::calibrateCleanPolars() {
    myRuntimeInfo->out << "Apply calibration factors for clean polars ..." << endl;
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        for (size_t machID(0); machID < polarDataTrimMap.at(CGID).cleanPolars.size(); ++machID) {
            for (size_t pointID(0); pointID < polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.size(); ++pointID) {
                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD =
                                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD * config_->factorDragCleanPolar.value();
                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD += config_->deltaTotalDragCleanPolar.value();
            }
        }
    }
}

void tandemDefaultStrategy::calibrateHighLiftPolars() {
    myRuntimeInfo->out << "Apply calibration factors for high lift polars ..." << endl;
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
            for (size_t pointID(0); pointID < polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.size(); ++pointID) {
                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD =
                                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD * config_->factorDragHighLiftPolar.value();
                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD += config_->deltaTotalDragHighLift.value();
            }
        }
    }
}

void tandemDefaultStrategy::writePolarXML() {
    myRuntimeInfo->out << "Create XML-Polar file ..." << endl;
    // Construct the file path and convert to absolute path
    std::filesystem::path filePath = std::filesystem::absolute(this->rtIO_->getAeroDataDir() + "/");
    // Ensure the directory exists
    if (!std::filesystem::exists(filePath.parent_path())) {
        std::filesystem::create_directories(filePath.parent_path());
    }
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        polarXMLFilePath = rtIO_->getAeroDataDir() + "/";
        polarXMLFilePath += rtIO_->aircraft_model();
        if (!config_->doTrim.value()) {
            this->polarXMLFilePath += "_polar.xml";
        } else {
            this->polarXMLFilePath += "_" + config_->CGPositions.at(CGID) + "_trimmed_polar.xml";
        }
        // file name for design case and sizing loop
        if (CGID == 0) {
            this->designPolarXMLFilePath = polarXMLFilePath;
            this->designPolarXMLFilePath = designPolarXMLFilePath.substr(designPolarXMLFilePath.find_last_of("/") + 1);
        }
        ofstream XML;
        XML.open(polarXMLFilePath.c_str());
        if (!XML) {
            myRuntimeInfo->err << polarXMLFilePath + " could not be opened!" << endl;
            exit(1);
        } else {
            XML << "<?xml version=\"1.0\"?>" << endl;
            XML << "<polar-exchange-file version=\"1.0\">" << endl;
            XML << "    <comment>The 'PEF-Browser' application, a polar file converter is available in the software section at:";
            XML << " http://www.MH-AeroTools.de/airfoils/</comment>" << endl;
            XML << "    <liftingLineVersion>" << LILI_Version << "</liftingLineVersion>" << endl;
            XML << "    <dragComponentMethods>" << endl;
            XML << "        <inducedDrag>Lifting Line + Without Corrections for fuselage and nacelles</inducedDrag>" << endl;
            XML << "        <viscousDrag>" << "Raymer" << "</viscousDrag>" << endl;
            XML << "        <waveDrag>" << "Mason" << "</waveDrag>" << endl;
            XML << "    </dragComponentMethods>" << endl;
            XML << "    <configurations>" << endl;
            XML << "        <name>" << rtIO_->aircraft_model() << "</name>" << endl;
            XML << "        <comment>Polars for wing configuration</comment>" << endl;
            XML << "        <configuration ID=\"clean\">" << endl;
            XML << "            <name>Clean</name>" << endl;
            XML << "            <comment>Clean Configuration</comment>" << endl;
            XML << "            <polars>" << endl;
            for (size_t MachID(0); MachID < polarDataTrimMap.at(CGID).cleanPolars.size(); MachID++) {
                XML << "                <polar ID=\"" << MachID + 1 << "\" ExtrapolationMargin=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).extrapolationMargin
                                    << "\" AllowGridChange=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).allowGridChange <<"\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cd_turb_str(""), Cm_str(""), origin_polar_str(""), i_backWing_str("");
                for (size_t i(0); i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size(); ++i) {
                    Alpha_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).AoA);
                    Cl_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CL);
                    Cd_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cd_turb_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cm_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CM);
                    origin_polar_str += "seg0-0.0";
                    i_backWing_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).iStabPolar);
                    if (i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cd_turb_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_backWing_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cd_turb>" << Cd_turb_str << "</Cd_turb>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iBackWing_polar>"<< i_backWing_str << "</iBackWing_polar>" << endl;
                XML << "                </polar>" << endl;
            }
            XML << "            </polars>" << endl;
            XML << "        </configuration>" << endl;
            for (const pair<const string, tandemDefaultPolar>& thePolarIterator : polarDataTrimMap.at(CGID).highLiftPolars) {
                XML << "        <configuration ID=\"" << thePolarIterator.first << "\">" << endl;
                XML << "            <name>" << thePolarIterator.first << "</name>" << endl;
                XML << "            <comment>" << thePolarIterator.first << " Configuration</comment>" << endl;
                XML << "            <polars>" << endl;
                // Inner config loop: Since the myAcftPt->highLiftPolar elements are ordered by config name, configs can be gathered here until the next one appears.
                XML << "                <polar ID=\"" << num2Str(1) << "\" ExtrapolationMargin =\"" << thePolarIterator.second.extrapolationMargin
                                        << "\" AllowGridChange=\"" << thePolarIterator.second.allowGridChange << "\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" << thePolarIterator.second.ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" << thePolarIterator.second.altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" <<thePolarIterator.second.MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cm_str(""), origin_polar_str(""), i_backWing_str("");
                for (size_t j(0); j < thePolarIterator.second.Points.size(); ++j) {
                    Alpha_str += num2Str(thePolarIterator.second.Points.at(j).AoA);
                    Cl_str += num2Str(thePolarIterator.second.Points.at(j).CL);
                    Cd_str += num2Str(thePolarIterator.second.Points.at(j).CD);
                    Cm_str += num2Str(thePolarIterator.second.Points.at(j).CM);
                    origin_polar_str += "seg0-0.0";
                    i_backWing_str += num2Str(thePolarIterator.second.Points.at(j).iStabPolar);
                    if (j < thePolarIterator.second.Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_backWing_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iBackWing_polar>"<< i_backWing_str << "</iBackWing_polar>" << endl;
                XML << "                </polar>" << endl;
                XML << "            </polars>" << endl;
                XML << "        </configuration>" << endl;
                }
            XML << "    </configurations>" << endl;
            XML << "</polar-exchange-file>" << endl;
        }
        XML.close();
    }
}

void tandemDefaultStrategy::writePolarData() {
     ofstream plot;
    /* plot all untrimmed polars */
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (const pair<const double, PolarData>& thePolarIterator : polarDataCGMap.at(CGID)) {
            /* plot all clean polars */
            for (const pair<const int, tandemDefaultPolar>& theMachIDIterator : thePolarIterator.second.cleanPolars) {
                stringstream machNumber;
                stringstream iBackWingStream;
                machNumber << setprecision(2) << fixed << theMachIDIterator.second.MachNumber;
                iBackWingStream << setprecision(2) << fixed << Rounding(config_->backWingAngles.at(thePolarIterator.first), 2);
                std::string csvFiles = "tandemCalculatePolar_M" + replaceAll(machNumber.str(), ".", "") + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iBackWingStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFiles.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFiles << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << theMachIDIterator.second.MachNumber << endl;
                    plot << "AoA; C_L; C_Dtotal; C_Dind; C_Dvisc; C_Dwave; C_M; L/D; i_backWing; C_Dvisc_frontWing; "
                    << "C_Dwave_frontWing; C_Dind_frontWing; C_Dvisc_backWing; C_Dwave_backWing; C_Dind_backWing; C_L_frontWing; C_L_backWing" << endl;
                    for (size_t i(0); i < theMachIDIterator.second.Points.size(); ++i) {
                        plot << setprecision(3) << fixed << theMachIDIterator.second.Points.at(i).AoA << ";";
                        plot << setprecision(4) << fixed << theMachIDIterator.second.Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CM << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CL / theMachIDIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).iStabPolar << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsFrontWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsFrontWing.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsFrontWing.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsBackWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsBackWing.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsBackWing.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsFrontWing.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsBackWing.at(i).CL << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (const pair<const double, PolarData>& thePolarIterator : polarDataCGMap.at(CGID)) {
            /* plot all clean polars */
            for (const pair<const std::string, tandemDefaultPolar>& theConfigIterator : thePolarIterator.second.highLiftPolars) {
                stringstream machNumber;
                stringstream iBackWingStream;
                machNumber << setprecision(2) << fixed << theConfigIterator.second.MachNumber;
                iBackWingStream << setprecision(2) << fixed << Rounding(config_->backWingAngles.at(thePolarIterator.first), 2);
                std::string csvFiles = "tandemCalculatePolar_" + theConfigIterator.first + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iBackWingStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFiles.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFiles << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << theConfigIterator.second.MachNumber << endl;
                    plot << "AoA; C_L; C_Dtotal; C_M; L/D; i_backWing;" << endl;
                    for (size_t i(0); i < theConfigIterator.second.Points.size(); ++i) {
                        plot << setprecision(3) << fixed << theConfigIterator.second.Points.at(i).AoA << ";";
                        plot << setprecision(4) << fixed << theConfigIterator.second.Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CM << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CL / theConfigIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).iStabPolar << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
    size_t cruiseMachID(0);
    stringstream machNumber;
    machNumber << setprecision(2) << fixed << data_->initialCruiseMachNumber;
    for (size_t machID(0); machID < polarDataTrimMap.at(0).cleanPolars.size(); ++machID) {
        if (accuracyCheck(polarDataTrimMap.at(0).cleanPolars.at(machID).MachNumber, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            cruiseMachID = machID;
            break;
        }
    }
    std::string csvFiles = "tandemCalculatePolar_frontWingliftDistribution_" + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFiles.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFiles << endl;
        plot << "# (1)C_L; (2)C_L_frontwing; (3)y, m; (4)C_l; (5)Cl*l, m" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsFrontWing.size(); ++pointID) {
            vector <double> clDistribution_tmp = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsFrontWing.at(pointID).Cl_distr;
            vector <double> yStations_tmp      = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsFrontWing.at(pointID).yStations;
            reverse(clDistribution_tmp.begin(), clDistribution_tmp.end());
            reverse(yStations_tmp.begin(), yStations_tmp.end());
            // for Cl distr analysis at single CL
            double CL_tmp(0.51);
            if (true || fabs(polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL - CL_tmp) < 1e-5) {  // uncomment "true || " for analysis
                string delimiter(";");
                for (size_t j(0); j < clDistribution_tmp.size(); ++j) {
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << delimiter;
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsFrontWing.at(pointID).CL << delimiter;
                    plot << yStations_tmp.at(j) << delimiter;
                    plot << clDistribution_tmp.at(j) << delimiter;
                    plot << getLocalChordLength(liftingSurfaces.at(0), yStations_tmp.at(j)) << endl;
                }
                plot << endl << endl;
            }
        }
        plot.close();
    }
    std::string csvFiles_back = "tandemCalculatePolar_backWingliftDistribution_" + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFiles_back.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFiles_back + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFiles_back << endl;
        plot << "# (1)C_L; (2)C_Lbackwing; (3)y, m; (4)C_l; (5)Cl*l, m" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsBackWing.size(); ++pointID) {
            vector <double> clDistribution_tmp = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsBackWing.at(pointID).Cl_distr;
            vector <double> yStations_tmp      = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsBackWing.at(pointID).yStations;
            reverse(clDistribution_tmp.begin(), clDistribution_tmp.end());
            reverse(yStations_tmp.begin(), yStations_tmp.end());
            // for Cl distr analysis at single CL
            double CL_tmp(0.51);
            if (true || fabs(polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL - CL_tmp) < 1e-5) {  // uncomment "true || " for analysis
                string delimiter(";");
                for (size_t j(0); j < clDistribution_tmp.size(); ++j) {
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << delimiter;
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsBackWing.at(pointID).CL << delimiter;
                    plot << yStations_tmp.at(j) << delimiter;
                    plot << clDistribution_tmp.at(j) << delimiter;
                    plot << getLocalChordLength(liftingSurfaces.at(1), yStations_tmp.at(j)) << endl;
                }
                plot << endl << endl;
            }
        }
        plot.close();
    }
    std::string csvFile = "tandem_trimmed_designCruisePolar_" + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFile + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFile << endl;
                            plot << "AoA; C_L; C_Dtotal; C_Dind; C_Dvisc; C_Dwave; C_M; L/D; i_backWing" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.size(); ++pointID) {
            plot << setprecision(3) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).AoA << ";";
            plot << setprecision(4) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_ind << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_visc << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_wave << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CM << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL
                                        / polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).iStabPolar << ";";
            plot << endl;
        }
        plot.close();
    }
}

void tandemDefaultStrategy::writeViscDragData() {
    ofstream plot;
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (size_t backWingID(0); backWingID < polarDataCGMap.at(CGID).size(); ++backWingID) {
            for (size_t machID(0); machID < polarDataCGMap.at(CGID).at(backWingID).cleanPolars.size(); ++machID) {
                stringstream machNumber;
                stringstream iBackWingStream;
                machNumber << setprecision(2) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).MachNumber;
                iBackWingStream << setprecision(2) << fixed << Rounding(config_->backWingAngles.at(backWingID), 2);
                std::string csvFile = "tandem_viscDrag_M" + replaceAll(machNumber.str(), ".", "") + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iBackWingStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFile.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFile + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFile << endl;
                    plot << "# Reynolds Number" << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).ReynoldsNumber << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).MachNumber << endl;
                    plot << "AoA; C_L; C_D_visc; C_D_visc_frontWing; C_D_visc_backWing; C_D_visc_Fin; C_D_visc_Fuselage; C_D_visc_Nacelles; C_D_visc_Misc;" << endl;
                    for (size_t i(0); i < polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).Points.size(); ++i) {
                        plot << setprecision(3) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).Points.at(i).AoA << ";";
                        plot << setprecision(3) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).Points.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsFrontWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsBackWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsFin.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsFuselage.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsNacelles.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(backWingID).cleanPolars.at(machID).PointsMisc.at(i).CD_visc << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
}

void tandemDefaultStrategy::setReferenceValues() {
    data_->span.set_value(getHalfSpan(liftingSurfaces.at(0)));
    data_->MAC.set_value(geom2::measure::mean_aerodynamic_chord(liftingSurfaces.at(0)));
    data_->S_ref.set_value(geom2::measure::reference_area(liftingSurfaces.at(0)) + geom2::measure::reference_area(liftingSurfaces.at(1)));
    data_->neutral_point_xCoordinate.set_value(getNeutralPointXposition(getTrimmedTandemCruisePolar(polarDataCGMap.at(0).at(0).cleanPolars, data_->initialCruiseMachNumber),
                                                                        data_->MAC.value(), data_->CoGPositions["design"].xCoordinate));
}

void tandemDefaultStrategy::setOutputValues() {
    data_->C_L_optimumCruise.set_value(getCLoptCruise());
    data_->CL_max_landing.set_value(getMaxCL(polarDataTrimMap.at(0).highLiftPolars.at("landing")));
    data_->C_LmaxTO.set_value(getMaxCL(polarDataTrimMap.at(0).highLiftPolars.at("takeoff")));
    data_->C_LgroundRoll.set_value(0.4);
    data_->max_spoiler_factor.set_value(2.5);
    // handle the axcml entries
    data_->polar_file.set_value(this->designPolarXMLFilePath);
    data_->polar_file.update(rtIO_->acxml);
    vector<std::string>types{"cruise", "departure", "departure", "departure", "approach", "approach", "approach"};
    vector<std::string>names{"clean", "takeoff", "takeoff_landing_gear_retracted", "climb", "approach", "approach_landing_gear_out", "landing"};
    //Write aerodynamic node with description and tool level
    rtIO_->acxml.at("analysis/aerodynamics").addAttrib("description", "Aerodynamical analysis");
    rtIO_->acxml.at("analysis/aerodynamics").addAttrib("tool_level", EndnodeReadOnly<std::string>("control_settings/own_tool_level").read(rtIO_->moduleConfig).value());
    std::string subPath = "analysis/aerodynamics/polar";
    Endnode<int> configurations(Endnode<int>(subPath + "/configurations", "Number of configurations in the polar file", types.size()));
    configurations.update(rtIO_->acxml);
    // setup configuration nodes
    for (size_t ID = 0; ID < types.size(); ID++) {
        rtIO_->acxml.at(subPath)["configuration@" + num2Str(ID)];
        rtIO_->acxml.at(subPath + "/configuration@" + num2Str(ID)).addAttrib("description", "Configuration in polar file marked with ID - name can vary");
        Endnode<string>type(Endnode<string>(subPath + "/configuration@" + num2Str(ID) + "/type", " Type", "cruise"));
        Endnode<string>name(Endnode<string>(subPath + "/configuration@" + num2Str(ID) + "/name", " Name", "clean"));
        type.set_value(types.at(ID));
        name.set_value(names.at(ID));
        type.update(rtIO_->acxml);
        name.update(rtIO_->acxml);
    }
}

void tandemDefaultStrategy::update_xml_entries(std::string mode) {
    data_->updateNodes(mode);
}

double tandemDefaultStrategy::getCLoptCruise() {
    double LoverDmax(0.0);
    double CLopt(0.0);
    std::vector<defaultPolarPoint> theTrimmedCruisePolar(getTrimmedTandemCruisePolar(polarDataTrimMap.at(0).cleanPolars, data_->initialCruiseMachNumber));
    for (size_t pointID(0); pointID < theTrimmedCruisePolar.size(); ++pointID) {
        double LoverD = theTrimmedCruisePolar.at(pointID).CL / theTrimmedCruisePolar.at(pointID).CD;
            if (LoverD > LoverDmax) {
                LoverDmax = LoverD;
                CLopt = theTrimmedCruisePolar.at(pointID).CL;
            }
    }
    return CLopt;
}

double tandemDefaultStrategy::getMaxCL(const tandemDefaultPolar& aPolar) {
    double maxCL = -10000;
    for (size_t ID = 0; ID < aPolar.Points.size(); ++ID) {
        if (aPolar.Points.at(ID).CL > maxCL) {
            maxCL = aPolar.Points.at(ID).CL;
        }
    }
    return maxCL;
}
