/**
 * \file cantileverTandemWingDesignReport.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Report for cantilever tandem wing design
 * \version 0.1
 * \date 2024-02-14
 *
 * @copyright Copyright (c) 2024
 *
 */
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <moduleBasics/html.h>

#include "cantileverTandemWingDesign.h"

namespace tandem {
namespace cantilever {
void Wing::set_html_body() {
  /* Add box data */
  reporter.htmlReportStream() << "<div class=\"box data\">\n";
  /* Add headline Data*/
  reporter.htmlReportStream() << "<h2>Data</h2>\n";

/* Table 1 - Positioning paramters */

  reporter.htmlReportStream()
  << "<table class=\"content-table\">\n"
  << "<caption>Positioning parameters of tandem wing set</caption>\n"
  << "<thead>\n"
  << "<tr>\n"
  << "<th>parameter</th><th>symbol</th><th>unit</th><th>value</th>\n"
  << "</tr>\n"
  << "</thead>\n"
  << "<tbody>\n"
  << "<tr>\n"
  << std::format("<td>Horizontal LE position of front wing</td><td>x<sub>front,LE</sub></td><td>m</td><td>{:.4f}</td>\n",
                 (data->wing_position.x.value() + data->front_wing.origin.x()))
  << "</tr>\n"
  << std::format("<td>Horizontal LE position of back wing</td><td>x<sub>back,LE</sub></td><td>m</td><td>{:.4f}</td>\n",
                 (data->wing_position.x.value() + data->back_wing.origin.x()))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Verical LE position of front wing</td><td>z<sub>front,LE</sub></td><td>m</td><td>{:.4f}</td>\n",
                 (data->wing_position.z.value() + data->front_wing.origin.z()))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Verical LE position of back wing</td><td>z<sub>back,LE</sub></td><td>m</td><td>{:.4f}</td>\n",
                 (data->wing_position.z.value() + data->back_wing.origin.z()))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Leading edge stagger</td><td>St<sub>LE</sub></td><td>m</td><td>{:.4f}</td>\n",
                 (data->back_wing.origin.x() - data->front_wing.origin.x()))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Relative leading edge stagger</td><td>St<sub>rel,LE</sub></td><td>1</td><td>{:.4f}</td>\n",
                 (data->back_wing.origin.x() - data->front_wing.origin.x()) / geom2::measure::length(data->fuselage))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>1/4 MAC stagger</td><td>St<sub>25_MAC</sub></td><td>m</td><td>{:.4f}</td>\n",
                ((data->back_wing.origin.x()
                + geom2::measure::offset_LE(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing)).x() 
                + 0.25 * geom2::measure::chord(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))
                * cos(geom2::measure::twist(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))))
                - (data->front_wing.origin.x()
                + geom2::measure::offset_LE(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing)).x() 
                + 0.25 * geom2::measure::chord(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))
                * cos(geom2::measure::twist(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))))) 
                )  
                            
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Relative 1/4 MAC stagger</td><td>St<sub>rel,25_MAC</sub></td><td>1</td><td>{:.4f}</td>\n",
                ((data->back_wing.origin.x()
                + geom2::measure::offset_LE(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing)).x() 
                + 0.25 * geom2::measure::chord(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))
                * cos(geom2::measure::twist(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))))
                - (data->front_wing.origin.x()
                + geom2::measure::offset_LE(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing)).x() 
                + 0.25 * geom2::measure::chord(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))
                * cos(geom2::measure::twist(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))))) 
                / geom2::measure::length(data->fuselage))
                            
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Vertical leading edge gap</td><td>G<sub>LE</sub></td><td>m</td><td>{:.4f}</td>\n", 
                data->back_wing.origin.z() - data->front_wing.origin.z())
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Relative vertical leading edge gap</td><td>G<sub>rel,LE</sub></td><td>1</td><td>{:.4f}</td>\n", 
                (data->back_wing.origin.z() - data->front_wing.origin.z()) / geom2::measure::height_max(data->fuselage))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Vertical 1/4 MAC gap</td><td>G<sub>25_MAC</sub></td><td>m</td><td>{:.4f}</td>\n", 
                ((data->back_wing.origin.z()
                + geom2::measure::offset_LE(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing)).y() 
                + 0.25 * geom2::measure::chord(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))
                * sin(geom2::measure::twist(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))))
                - (data->front_wing.origin.z()
                + geom2::measure::offset_LE(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing)).y() 
                + 0.25 * geom2::measure::chord(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))
                * sin(geom2::measure::twist(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))))) 
                )  
  << "</tr>\n"

  << "<tr>\n"
  << std::format("<td>Relative vertical 1/4 MAC gap</td><td>G<sub>rel,25_MAC</sub></td><td>1</td><td>{:.4f}</td>\n", 
                ((data->back_wing.origin.z()
                + geom2::measure::offset_LE(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing)).y() 
                + 0.25 * geom2::measure::chord(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))
                * sin(geom2::measure::twist(data->back_wing, geom2::measure::mean_aerodynamic_chord_position(data->back_wing))))
                - (data->front_wing.origin.z()
                + geom2::measure::offset_LE(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing)).y() 
                + 0.25 * geom2::measure::chord(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))
                * sin(geom2::measure::twist(data->front_wing, geom2::measure::mean_aerodynamic_chord_position(data->front_wing))))) 
                / geom2::measure::height_max(data->fuselage))
  << "</tr>\n"

  << "<tr>\n"
  << std::format("<td>Fuselage length</td><td>l<sub>fus</sub></td><td>m</td><td>{:.4f}</td>\n",
                 geom2::measure::length(data->fuselage))
  << "</tr>\n"
  << "<tr>\n"
  << std::format("<td>Fuselage max. height</td><td>h<sub>fus,max</sub></td><td>m</td><td>{:.4f}</td>\n",
                 geom2::measure::height_max(data->fuselage))
  << "</tr>\n"
  << "</tbody>\n"
  << "</table>\n";
  reporter.htmlReportStream()
  << "<p style=\"font-size: 14px; font-style: italic;\">"
  << "Note: Horizontal relative values are calculated as fractions of the fuselage length. <br>"
  << "Vertical relative values are calculated as fractions of the maximum fuselage height."
  << "</p>\n";



  /* Table 2 - General parameters for Front Wing*/
  reporter.htmlReportStream()
      << "<table class=\"content-table\">\n"
      << "<caption>General parameters of front wing</caption>\n"
      << "<thead>\n"
      << "<tr>\n"
      << "<th>parameter</th><th>symbol</th><th>unit</th><th>value</th>\n"
      << "</tr>\n"
      << "</thead>\n"
      << "<tbody>\n"
      << "<tr>\n"
      << std::format("<td>Reference area</td><td>S<sub>ref</sub></td><td>m<sup>2</sup></td><td>{:.2f}</td>\n",
                     geom2::measure::reference_area(data->front_wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Mean aero. chord</td><td>MAC</td><td>m</td><td>{:.2f}</td>\n",
                     geom2::measure::mean_aerodynamic_chord(data->front_wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Span</td><td>b</td><td>m</td><td>{:.2f}</td>\n", geom2::measure::span(data->front_wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Aspect ratio</td><td>AR</td><td>1</td><td>{:.2f}</td>\n",
                     geom2::measure::aspect_ratio(data->front_wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Taper ratio</td><td>&lambda;</td><td>1</td><td>{:.2f}</td>\n",
                     geom2::measure::taper_ratio(data->front_wing))
      << "</tr>\n"
      << "</tbody>\n"
      << "</table>\n";

/* Table 3 - General parameters for Back Wing */
  reporter.htmlReportStream()
    << "<table class=\"content-table\">\n"
    << "<caption>General parameters of back wing</caption>\n"
    << "<thead>\n"
    << "<tr>\n"
    << "<th>parameter</th><th>symbol</th><th>unit</th><th>value</th>\n"
    << "</tr>\n"
    << "</thead>\n"
    << "<tbody>\n"
    << "<tr>\n"
    << std::format("<td>Reference area</td><td>S<sub>ref</sub></td><td>m<sup>2</sup></td><td>{:.2f}</td>\n",
                   geom2::measure::reference_area(data->back_wing))
    << "</tr>\n"
    << "<tr>\n"
    << std::format("<td>Mean aero. chord</td><td>MAC</td><td>m</td><td>{:.2f}</td>\n",
                   geom2::measure::mean_aerodynamic_chord(data->back_wing))
    << "</tr>\n"
    << "<tr>\n"
    << std::format("<td>Span</td><td>b</td><td>m</td><td>{:.2f}</td>\n", geom2::measure::span(data->back_wing))
    << "</tr>\n"
    << "<tr>\n"
    << std::format("<td>Aspect ratio</td><td>AR</td><td>1</td><td>{:.2f}</td>\n",
                   geom2::measure::aspect_ratio(data->back_wing))
    << "</tr>\n"
    << "<tr>\n"
    << std::format("<td>Taper ratio</td><td>&lambda;</td><td>1</td><td>{:.2f}</td>\n",
                   geom2::measure::taper_ratio(data->back_wing))
    << "</tr>\n"
    << "</tbody>\n"
    << "</table>\n";

  /* Table 4 - Section parameters of Front Wing*/
  reporter.htmlReportStream() << "<table class=\"content-table\">\n"
                              << "<caption>Section parameters of front wing</caption>\n"
                              << "<thead>\n"
                              << "<tr>\n"
                              << "<th>parameter</th><th>symbol</th><th>unit</th>";
  for (size_t i = 0; i < data->front_wing.sections.size(); ++i) {
    reporter.htmlReportStream() << std::format("<th>S<sub>{:02d}</sub></th>", i);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</thead>\n"
                              << "<tbody>\n"
                              /* Dimensionless half span - eta*/
                              << "<tr>\n"
                              << "<td>Dimensionless half span</td><td>&eta;</td><td>1</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>",
                                               -2 * section.origin.z() / geom2::measure::span(data->front_wing));
  }
  /* Spanwise coordinate - y*/
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "<td>Spanwise coordinate</td><td>y</td><td>m</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", -section.origin.z());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Chord - c*/
                              << "<tr>\n"
                              << "<td>Chord</td><td>c</td><td>m</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Leading edge - phi_LE*/
                              << "<tr>\n"
                              << "<td>Sweep leading edge</td><td>&phi;<sub>LE</sub></td><td>&deg;</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->front_wing, section.origin.z(), 0.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - 25 percent - phi_25%*/
                              << "<tr>\n"
                              << "<td>Sweep quarter chord</td><td>&phi;<sub>25</sub></td><td>&deg;</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->front_wing, section.origin.z(), 0.25));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Trailing edge percent - phi_TE*/
                              << "<tr>\n"
                              << "<td>Sweep Trailing edge</td><td>&phi;<sub>TE</sub></td><td>&deg;</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->front_wing, section.origin.z(), 1.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Dihedral - nu*/
                              << "<tr>\n"
                              << "<td>Dihedral</td><td>&nu;</td><td>&deg;</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", geom2::detail::to_degrees * geom2::measure::dihedral(data->front_wing, section.origin.z()));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Twist - epsilon*/
                              << "<tr>\n"
                              << "<td>Twist angle</td><td>&epsilon;</td><td>&deg;</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", geom2::detail::to_degrees * section.rotation_z);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Thickness to chord - epsilon*/
                              << "<tr>\n"
                              << "<td>Thickness ratio</td><td>t/c</td><td>%</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", 100 * geom2::measure::thickness_max(section) / section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Airfoil - none */
                              << "<tr>\n"
                              << "<td>Airfoil</td><td>-</td><td>-</td>";
  for (auto section : data->front_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{}</td>", section.name);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</tbody>\n"
                              << "</table>\n";

  /* Table 2 - Section parameters of Back Wing*/
  reporter.htmlReportStream() << "<table class=\"content-table\">\n"
                              << "<caption>Section parameters of back wing</caption>\n"
                              << "<thead>\n"
                              << "<tr>\n"
                              << "<th>parameter</th><th>symbol</th><th>unit</th>";
  for (size_t i = 0; i < data->back_wing.sections.size(); ++i) {
    reporter.htmlReportStream() << std::format("<th>S<sub>{:02d}</sub></th>", i);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</thead>\n"
                              << "<tbody>\n"
                              /* Dimensionless half span - eta*/
                              << "<tr>\n"
                              << "<td>Dimensionless half span</td><td>&eta;</td><td>1</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>",
                                               -2 * section.origin.z() / geom2::measure::span(data->back_wing));
  }
  /* Spanwise coordinate - y*/
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "<td>Spanwise coordinate</td><td>y</td><td>m</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", -section.origin.z());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Chord - c*/
                              << "<tr>\n"
                              << "<td>Chord</td><td>c</td><td>m</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Leading edge - phi_LE*/
                              << "<tr>\n"
                              << "<td>Sweep leading edge</td><td>&phi;<sub>LE</sub></td><td>&deg;</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->back_wing, section.origin.z(), 0.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - 25 percent - phi_25%*/
                              << "<tr>\n"
                              << "<td>Sweep quarter chord</td><td>&phi;<sub>25</sub></td><td>&deg;</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->back_wing, section.origin.z(), 0.25));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Trailing edge percent - phi_TE*/
                              << "<tr>\n"
                              << "<td>Sweep Trailing edge</td><td>&phi;<sub>TE</sub></td><td>&deg;</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->back_wing, section.origin.z(), 1.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Dihedral - nu*/
                              << "<tr>\n"
                              << "<td>Dihedral</td><td>&nu;</td><td>&deg;</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", geom2::detail::to_degrees * geom2::measure::dihedral(data->back_wing, section.origin.z()));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Twist - epsilon*/
                              << "<tr>\n"
                              << "<td>Twist angle</td><td>&epsilon;</td><td>&deg;</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", geom2::detail::to_degrees * section.rotation_z);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Thickness to chord - epsilon*/
                              << "<tr>\n"
                              << "<td>Thickness ratio</td><td>t/c</td><td>%</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", 100 * geom2::measure::thickness_max(section) / section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Airfoil - none */
                              << "<tr>\n"
                              << "<td>Airfoil</td><td>-</td><td>-</td>";
  for (auto section : data->back_wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{}</td>", section.name);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</tbody>\n"
                              << "</table>\n";


  /* Table 5 - Spar parameters of Front Wing*/
  for (size_t spar_id = 0; spar_id < data->front_spars.size(); ++spar_id) {
    reporter.htmlReportStream() << "<table class=\"content-table\">\n";
    if (spar_id == 0) {
      reporter.htmlReportStream() << "<caption>Spar parameters of front wing</caption>\n";
    }
    reporter.htmlReportStream() << "<thead>\n"
                                << "<tr>\n"
                                << std::format("<th>{:<10}</th><th>&eta; [1]<th>x/c [%]</th>",
                                               data->front_spars.at(spar_id).name)
                                << "\n</tr>\n"
                                << "</thead>\n"
                                << "<tbody>\n";
    /* Spar name - eta, x/c*/
    for (size_t i = 0; i < data->front_spars.at(spar_id).sections.size(); ++i) {
      reporter.htmlReportStream() << "<tr>\n";
      std::string tag = "";
      if (i == 0) {
        tag = "from";
      } else if (i == data->front_spars.at(spar_id).sections.size() - 1) {
        tag = "to";
      }
      reporter.htmlReportStream() << std::format("<td>{}</td><td>{:.2f}</td><td>{:.2f}</td>\n", tag,
                                                 data->front_spars.at(spar_id).sections.at(i).origin.z(),
                                                 data->front_spars.at(spar_id).sections.at(i).get_contour().vertex(0).x());
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  /* Table 6 - Spar parameters of Back Wing*/
  for (size_t spar_id = 0; spar_id < data->back_spars.size(); ++spar_id) {
    reporter.htmlReportStream() << "<table class=\"content-table\">\n";
    if (spar_id == 0) {
      reporter.htmlReportStream() << "<caption>Spar parameters of back wing</caption>\n";
    }
    reporter.htmlReportStream() << "<thead>\n"
                                << "<tr>\n"
                                << std::format("<th>{:<10}</th><th>&eta; [1]<th>x/c [%]</th>",
                                               data->back_spars.at(spar_id).name)
                                << "\n</tr>\n"
                                << "</thead>\n"
                                << "<tbody>\n";
    /* Spar name - eta, x/c*/
    for (size_t i = 0; i < data->back_spars.at(spar_id).sections.size(); ++i) {
      reporter.htmlReportStream() << "<tr>\n";
      std::string tag = "";
      if (i == 0) {
        tag = "from";
      } else if (i == data->back_spars.at(spar_id).sections.size() - 1) {
        tag = "to";
      }
      reporter.htmlReportStream() << std::format("<td>{}</td><td>{:.2f}</td><td>{:.2f}</td>\n", tag,
                                                 data->back_spars.at(spar_id).sections.at(i).origin.z(),
                                                 data->back_spars.at(spar_id).sections.at(i).get_contour().vertex(0).x());
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  /* Sort devices to leading, trailing and other devices for front wing*/
  /* device data - name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
   * rel_chord_to_end */
  using front_device_data = std::tuple<std::string, double, double, double, double, double, double>;
  std::vector<front_device_data> front_leading_edge_devices;
  std::vector<front_device_data> front_trailing_edge_devices;
  std::vector<front_device_data> front_other_devices;
  for (auto device : data->front_control_devices) {
    /* Device has only two sections (from + to) spanwise */
    double eta_from = device.sections.at(0).origin.z();
    double rel_chord_from_start = device.sections.at(0).get_contour().vertex(0).x(),
           rel_chord_from_end = device.sections.at(0).get_contour().vertex(1).x();
    double eta_to = device.sections.at(1).origin.z();
    double rel_chord_to_start = device.sections.at(1).get_contour().vertex(0).x(),
           rel_chord_to_end = device.sections.at(1).get_contour().vertex(1).x();

    if (std::fabs(rel_chord_from_start - rel_chord_to_start) < ACCURACY_MEDIUM &&
        std::fabs(rel_chord_from_start) < ACCURACY_MEDIUM) {
      front_leading_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                      rel_chord_to_start, rel_chord_to_end});
    } else if (std::fabs(rel_chord_from_end - rel_chord_to_end) < ACCURACY_MEDIUM &&
               std::fabs(rel_chord_from_end - 1) < ACCURACY_MEDIUM) {
      front_trailing_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                       rel_chord_to_start, rel_chord_to_end});
    } else {
      front_other_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                               rel_chord_to_start, rel_chord_to_end});
    }
  }

  /* Leading edge table*/
  if (!front_leading_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Leading edge control devices of front wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : front_leading_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!front_trailing_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Trailing edge control devices of front wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : front_trailing_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!front_other_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Other control devices of front wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : front_other_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

    /* Sort devices to leading, trailing and other devices for back wing*/
  /* device data - name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
   * rel_chord_to_end */
  using back_device_data = std::tuple<std::string, double, double, double, double, double, double>;
  std::vector<back_device_data> back_leading_edge_devices;
  std::vector<back_device_data> back_trailing_edge_devices;
  std::vector<back_device_data> back_other_devices;
  for (auto device : data->back_control_devices) {
    /* Device has only two sections (from + to) spanwise */
    double eta_from = device.sections.at(0).origin.z();
    double rel_chord_from_start = device.sections.at(0).get_contour().vertex(0).x(),
           rel_chord_from_end = device.sections.at(0).get_contour().vertex(1).x();
    double eta_to = device.sections.at(1).origin.z();
    double rel_chord_to_start = device.sections.at(1).get_contour().vertex(0).x(),
           rel_chord_to_end = device.sections.at(1).get_contour().vertex(1).x();

    if (std::fabs(rel_chord_from_start - rel_chord_to_start) < ACCURACY_MEDIUM &&
        std::fabs(rel_chord_from_start) < ACCURACY_MEDIUM) {
      back_leading_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                      rel_chord_to_start, rel_chord_to_end});
    } else if (std::fabs(rel_chord_from_end - rel_chord_to_end) < ACCURACY_MEDIUM &&
               std::fabs(rel_chord_from_end - 1) < ACCURACY_MEDIUM) {
      back_trailing_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                       rel_chord_to_start, rel_chord_to_end});
    } else {
      back_other_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                               rel_chord_to_start, rel_chord_to_end});
    }
  }

  /* Leading edge table*/
  if (!back_leading_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Leading edge control devices of back wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : back_leading_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!back_trailing_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Trailing edge control devices of back wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : back_trailing_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!back_other_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Other control devices of back wing</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : back_other_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  reporter.htmlReportStream() << "<table class=\"content-table\">\n"
                              << "<caption>Mass and CoG of tandem wing set</caption>\n"
                              << "<thead>\n"
                              << "<tr>\n"
                              << "<th>Mass</th><th>CoG<sub>x</sub></th><th>CoG<sub>y</sub></th><th>CoG<sub>z</sub></th>\n"
                              << "</tr>\n"
                              << "</thead>\n"
                              << "<tbody>\n"
                              << "<tr>\n"
                              << std::format("<td>{:.2f}kg</td><td>{:.2f}m</td><td>{:.2f}m</td><td>{:.2f}m</td>",
                                             data->wing_mass_properties.data["mass"].value(),
                                             data->wing_mass_properties.data["x"].value(),
                                             data->wing_mass_properties.data["y"].value(),
                                             data->wing_mass_properties.data["z"].value())
                              << "\n<tr>\n"
                              << "</tbody>\n"
                              << "</table>\n";
  /* close box data div */
  reporter.htmlReportStream() << "</div>\n";

  /* open box plot div */
  reporter.htmlReportStream() << "<div class=\"box plot\">\n";
  /* Add headline Data*/
  reporter.htmlReportStream() << "<h2>Plots</h2>\n";

  reporter.htmlReportStream() << "<h3>Geometry of front wing</h3>\n";
  auto front_planform_plot_path = plot_front_wing_planform();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n",front_planform_plot_path.string());

  reporter.htmlReportStream() << "<h3>Geometry of back wing</h3>\n";
  auto back_planform_plot_path = plot_back_wing_planform();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n",back_planform_plot_path.string());

  auto front_airfoil_plot_paths = plot_front_airfoils();
  reporter.htmlReportStream() << "<h3>Airfoils of front wing</h3>\n";
  for( auto airfoil_path : front_airfoil_plot_paths) {
    reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", airfoil_path.string());
  }

  auto back_airfoil_plot_paths = plot_back_airfoils();
  reporter.htmlReportStream() << "<h3>Airfoils of back wing</h3>\n";
  for( auto airfoil_path : back_airfoil_plot_paths) {
    reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", airfoil_path.string());
  }

  reporter.htmlReportStream() << "<h3>Thickness & Geometric Twist of front wing</h3>\n";
  auto front_thickness_and_twist_plot_path = plot_front_thickness_and_twist_distribution();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", front_thickness_and_twist_plot_path.string());

  reporter.htmlReportStream() << "<h3>Thickness & Geometric Twist of back wing</h3>\n";
  auto back_thickness_and_twist_plot_path = plot_back_thickness_and_twist_distribution();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", back_thickness_and_twist_plot_path.string());

  /* close box plot */
  reporter.htmlReportStream() << "</div>\n";
}
}  // namespace cantilever
}  // namespace tandem
