/**
 * \file tandemWingDesignConfig.h
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Wing design configuration for tandem aircrafts
 * \version 0.1
 * \date 2023-12-14
 *
 * @copyright Copyright (c) 2023
 *
 */
#ifndef WINGDESIGN_SRC_TANDEM_TANDEMWINGDESIGNCONFIG_H_
#define WINGDESIGN_SRC_TANDEM_TANDEMWINGDESIGNCONFIG_H_

#include <aixml/endnode.h>

#include <map>
#include <set>

#include "lib/io_methods/controlDeviceIO.h"
#include "lib/io_methods/sparIO.h"

namespace tandem {
namespace cantilever {

#include <aixml/endnode.h>

class WingDesignConfig {
 public:
  WingDesignConfig();
  ~WingDesignConfig() = default;
  void read(const node& xml);

  /* Modes */
  EndnodeReadOnly<std::string> wing_design_mode;
  EndnodeReadOnly<std::string> wing_configuration_mode;
  EndnodeReadOnly<std::string> wing_mass_mode;


  /* Design */
  /* Data for method - wing area calculation */
  EndnodeReadOnly<std::string> total_wing_area_calculation_mode;
  EndnodeReadOnly<double> user_total_wing_area;
  
  EndnodeReadOnly<double> user_wing_area_ratio;
  EndnodeReadOnly<double> user_relative_LE_stagger;

  /* Data for method - quarter chord sweep calculation */
  EndnodeReadOnly<std::string> sweep_calculation_mode;
  EndnodeReadOnly<double> user_front_sweep_angle_at_quarter_chord;
  EndnodeReadOnly<double> front_korn_technology_factor;
  EndnodeReadOnly<double> front_delta_drag_divergence_to_mach_design;

  EndnodeReadOnly<double> user_back_sweep_angle_at_quarter_chord;
  EndnodeReadOnly<double> back_korn_technology_factor;
  EndnodeReadOnly<double> back_delta_drag_divergence_to_mach_design;

  /* Data for method - taper ratio calculation */
  EndnodeReadOnly<std::string> taper_ratio_calculation_mode;
  EndnodeReadOnly<double> user_front_taper_ratio;

  EndnodeReadOnly<double> user_back_taper_ratio;

  /* Data for method - aspect ratio calculation */
  EndnodeReadOnly<std::string> aspect_ratio_calculation_mode;
  EndnodeReadOnly<double> user_front_aspect_ratio;

  EndnodeReadOnly<double> user_back_aspect_ratio;

  /* Data for method - dihedral calculation*/
  EndnodeReadOnly<std::string> dihedral_calculation_mode;
  EndnodeReadOnly<double> user_front_dihedral;

  EndnodeReadOnly<double> user_back_dihedral;
    //TODO kayra: seperated mode switches after here
  EndnodeReadOnly<std::string> front_relative_kink_position_calculation_mode;
  EndnodeReadOnly<double> user_front_relative_kink;

  EndnodeReadOnly<std::string> back_relative_kink_position_calculation_mode;
  EndnodeReadOnly<double> user_back_relative_kink;

  EndnodeReadOnly<double> user_front_max_inner_trailing_edge_sweep;

  EndnodeReadOnly<double> user_back_max_inner_trailing_edge_sweep;

  EndnodeReadOnly<double> track_based_front_initial_relative_kink;
  EndnodeReadOnly<double> track_based_front_max_inner_trailing_edge_sweep;
  EndnodeReadOnly<double> track_based_back_initial_relative_kink;
  EndnodeReadOnly<double> track_based_back_max_inner_trailing_edge_sweep;

  /* Data for method - wing profile and thickness calculation */
  EndnodeReadOnly<std::string> front_wing_profile_and_thickness_calculation_mode;
  std::vector<EndnodeReadOnly<std::string>> user_front_wing_profiles;
  std::vector<EndnodeReadOnly<double>> user_front_thickness_to_chord;
  std::vector<EndnodeReadOnly<double>> user_front_eta;
  std::vector<std::tuple<double, double, std::string>> user_front_wing_profiles_and_thickness_vector;
  EndnodeReadOnly<std::string> front_torenbeek_jenkinson_wing_profile;
  EndnodeReadOnly<double> front_max_thickness_to_chord_ratio;
  EndnodeReadOnly<double> front_airfoil_critical_factor;

  EndnodeReadOnly<std::string> back_wing_profile_and_thickness_calculation_mode;
  std::vector<EndnodeReadOnly<std::string>> user_back_wing_profiles;
  std::vector<EndnodeReadOnly<double>> user_back_thickness_to_chord;
  std::vector<EndnodeReadOnly<double>> user_back_eta;
  std::vector<std::tuple<double, double, std::string>> user_back_wing_profiles_and_thickness_vector;
  EndnodeReadOnly<std::string> back_torenbeek_jenkinson_wing_profile;
  EndnodeReadOnly<double> back_max_thickness_to_chord_ratio;
  EndnodeReadOnly<double> back_airfoil_critical_factor;

  /* Mass */


  // flops mass method
  EndnodeReadOnly<double> front_flops_fstrt;
  EndnodeReadOnly<double> front_flops_faert;
  EndnodeReadOnly<double> front_flops_fcomp;
  EndnodeReadOnly<double> front_flops_technology_factor;

  EndnodeReadOnly<double> back_flops_fstrt;
  EndnodeReadOnly<double> back_flops_faert;
  EndnodeReadOnly<double> back_flops_fcomp;
  EndnodeReadOnly<double> back_flops_technology_factor;

  // chiozzotto wer mass method
  EndnodeReadOnly<double> front_chiozzotto_technology_factor;
  EndnodeReadOnly<std::string> front_chiozzotto_material;

  EndnodeReadOnly<double> back_chiozzotto_technology_factor;
  EndnodeReadOnly<std::string> back_chiozzotto_material;

  /* Data for method - control devices */
  EndnodeReadOnly<std::string> front_control_device_mode;
  std::vector<ControlDevice> user_defined_front_control_devices;
  EndnodeReadOnly<std::string> front_high_lift_device_type_leading_edge;
  EndnodeReadOnly<std::string> front_high_lift_device_type_trailing_edge;

  EndnodeReadOnly<std::string> back_control_device_mode;
  std::vector<ControlDevice> user_defined_back_control_devices;
  EndnodeReadOnly<std::string> back_high_lift_device_type_leading_edge;
  EndnodeReadOnly<std::string> back_high_lift_device_type_trailing_edge;

  /* Data for method - spars */
  EndnodeReadOnly<std::string> front_spars_mode;
  std::vector<Spar> user_defined_front_spars;
  
  EndnodeReadOnly<std::string> back_spars_mode;
  std::vector<Spar> user_defined_back_spars;

  /* Additional paths */
  EndnodeReadOnly<std::string> common_airfoil_data_path;

};
}  // namespace cantilever
}  // namespace tandem

#endif  // WINGDESIGN_SRC_TAW_TAWWINGDESIGNCONFIG_H_
