/**
 * \file tandemWingDesignData.h
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Wing design data for tandem aircrafts
 * \version 0.1
 * \date 2023-12-14
 *
 * @copyright Copyright (c) 2023
 *
 */
#ifndef WINGDESIGN_SRC_TANDEM_TANDEMWINGDESIGNDATA_H_
#define WINGDESIGN_SRC_TANDEM_TANDEMWINGDESIGNDATA_H_

#include <aircraftGeometry2/geometry/surface.h>
#include <aixml/endnode.h>

#include <iostream>

#include "lib/io_methods/massPropertiesIO.h"


#include "lib/io_methods/positionIO.h"
namespace tandem {
namespace cantilever {
class WingDesignData {
 public:
  WingDesignData();
  ~WingDesignData() = default;

  void read(const node& xml);
  void read_fuselage(std::filesystem::path acxml, std::filesystem::path geometry_data_dir);
  void read_wing_mounted_engine_information(const node& xml);

  void update(node& xml);

  void update_front_wing(node& xml);
  void update_back_wing(node& xml);

  void update_front_spars(node& xml);
  void update_back_spars(node& xml);

  void update_front_control_devices(node& xml);
  void update_back_control_devices(node& xml);

  EndnodeReadOnly<std::string> specification_icao_aerodrome_reference_code;
  EndnodeReadOnly<std::string> specification_front_wing_mounting;
  EndnodeReadOnly<std::string> specification_back_wing_mounting;
  EndnodeReadOnly<double> specification_longitudinal_position;
  EndnodeReadOnly<std::string> specification_landing_gear_mounting;
  EndnodeReadOnly<double> specification_maximum_operating_mach;
  EndnodeReadOnly<double> specification_design_mach;
  EndnodeReadOnly<double> specification_design_altitude;
  EndnodeReadOnly<double> specification_maximum_load_factor;
  EndnodeReadOnly<double> sizing_point_wing_loading;
  EndnodeReadOnly<double> track_based_front_relative_kink;
  EndnodeReadOnly<double> track_based_back_relative_kink;

  Endnode<double> spanwise_kink;
  size_t specification_number_of_wing_mounted_engines;     //TODO kayra: I got everything for engines exactly same to back wing
  std::vector<std::tuple<double,double>> wing_mounted_engine_data;
  bool engines_exist = true;

  MassPropertiesIO mtom_mass_properties;
  MassPropertiesIO wing_mass_properties;
  
  PositionIO wing_position;

  PositionIO center_of_gravity_mtom;
  geom2::MultisectionSurface<geom2::PolygonSection> fuselage;
  geom2::MultisectionSurface<geom2::AirfoilSection> front_wing;
  geom2::MultisectionSurface<geom2::AirfoilSection> back_wing;

  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> front_control_devices;
  std::vector<std::tuple<double,double>> front_control_devices_deflections;
  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> back_control_devices;
  std::vector<std::tuple<double,double>> back_control_devices_deflections;

  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> front_spars;
  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> back_spars;

};
}  // namespace cantilever
}  // namespace tandem

#endif  // WINGDESIGN_SRC_TAW_TAWWINGDESIGNDATA_H_
